;;;;;
;;;;; The file with data window definitions and
;;;;; operations.
;;;;;

;; a function that reads in an identification and returns it as a symbol
;;
(define (read-word input)
   (let loop ((l (list)))
      (let ((c (peek-char input)))
	 (if (and (not (eof-object? c)) (char-alphabetic? c)) ;>>>> should be more relaxed (to accept "id_2")
	     (loop (cons (read-char input) l))
	     (string->symbol (list->string (reverse l)))))))

;; a function that reads in a number (either integer or float point)
;;
(define (read-number input)
  (let ((period-occurs 0))
    (let loop ((num (list)))
      (let ((c (peek-char input)))
        (if (and (not (eof-object? c))
                 (or (char-numeric? c)
                     (and (char=? c #\.)
                          (zero? period-occurs))))
            (let ((ch (read-char input)))
              (if (char=? ch #\.) (set! period-occurs 1))
              (loop (cons ch num)))
            (string->number (list->string (reverse num))))))))


;; a function that reads a set of numbers as a sequence of digits and commas and returns a list
;; of these numbers
;;
(define (read-set input)
  (let loop ((l (list)))
    (let ((c (peek-char input)))
      (if (and (not (eof-object? c))
               (char-numeric? c))
          (let ((num (read-number input)))
            (let ((next-c (peek-char input)))
              (if (and (not (eof-object? next-c))
                       (char=? next-c #\,))
                  (begin
                    (read-char input)
                    (loop (cons num l)))
                  (reverse (cons num l)))))
          (reverse l)))))

;(define (read-set input)
;   (let outer ((l (list)))
;      (let inner ((elem (list)))
;	 (let ((c (peek-char input)))
;            (cond
;	       ((and (not (eof-object? c)) (char-numeric? c))    (inner (cons (read-char input) elem)))
;	       ((and (not (eof-object? c)) (char=? c #\,))       (read-char input)
;								 (outer (cons (string->number (apply string (reverse elem))) l)))
;	       (else (reverse (cons (string->number (apply string (reverse elem)))
;				    l))))))))

;; a function that skips tabulation and white spaces
;;
(define (skip-spaces input)
   (let loop ((c (peek-char input)))
      (if (and (not (eof-object? c))
	       (or (char=? c #\space) (char=? c #\tab)))
	  (begin
	     (read-char input)
	     (loop (peek-char input))))))

;; a function that reads a line of names
;;
(define (read-name-list input)
   (let loop ((c (peek-char input))(l (list)))
      (cond
	 ((eof-object? c)                           (reverse l))
	 ((char=? c #\newline)                      (begin (read-char input)
							   (reverse l)))
	 ((or (char=? c #\space) (char=? c #\tab))  (begin (skip-spaces input)
							   (loop (peek-char input) l)))
	 ((char-alphabetic? c)                      (let ((word (read-word input)))
						       (loop (peek-char input) (cons word l))))
	 (else (er "read-name-list" "expecting a set of words")
	       '*error-reading-sensor-names*))))

;; a function that reads a line of data sets and creates an association list
;; whose keys are names of variables
;;
(define (read-data-set-list input name-list)

  (define (unfold-data var arg)

    (define (vectorize len pos)
      (let loop ((i 0)(res '()))
	(cond ((= i len) res)
	      ((= i pos) (loop (+ i 1) (append res (list 1))))
	      (else      (loop (+ i 1) (append res (list 0)))))))

    (define (get-position-from-id l elem)
      (let loop ((l l)(i 0))
	(if (null? l)
	    #f
	    (if (eq? (car l) elem)
		i
		(loop (cdr l) (+ i 1))))))

    (let ((def (definitions-hash 'get-first-entry var)))
      (if (definition? def)
	  (cond 
	   ((number? arg) (let ((dim (definition-dim def)))
			    (if (< arg dim)
				(vectorize dim arg)
				#f)))
	   ((symbol? arg) (let* ((values (definition-values def))
				 (position (get-position-from-id values arg)))
			    (if position
				(vectorize (definition-dim def) position)
				#f)))
	   (else #f))
	  #f)))

  (let loop ((c (peek-char input))(l (list))(names name-list))
    (cond
     ((eof-object? c)                           (reverse l))
     ((char=? c #\newline)                      (read-char input)
                                                (reverse l))
     ((or (char=? c #\space) (char=? c #\tab))  (skip-spaces input)
                                                (loop (peek-char input) l names))
     ((char-numeric? c)                         (let ((new-set (read-set input)))
						  (if (= (length new-set) 1)
						      (let ((num (car new-set)))
							(if (= (truncate num) num)
							    (let* ((var (car names))
								   (unfolded-data (unfold-data var num)))
							      (loop (peek-char input)
								    (cons (cons var (list unfolded-data)) l)
								    (cdr names)))
							    (begin
							      (er "read-data-set-list" "the data dimension is not an integer")
							      '*error-read-data-set-list*)))
						      (loop (peek-char input) 
							    (cons (cons (car names) (list new-set)) l) 
							    (cdr names)))))
     ((char-alphabetic? c)                      (let* ((id (read-word input))
						       (var (car names))
						       (unfolded-data (unfold-data var id)))
						  (loop (peek-char input)
							(cons (cons var (list unfolded-data)) l)
							(cdr names))))
     (else (er "read-data-set-list" "expecting a set of numbers")
	   '*error-reading-data*))))


;; The structure definition for sensors.
;;
;; input-file -- an input port.
;; name-list  -- a list of names of the sensors.
;;
(define-structure sensors input-file name-list)

;; a function that retrieves a list of names of sensors for a sensors structure.
;;
(define (get-sensors-name-list sen)
   (if (sensors? sen)
       (sensors-name-list sen)
       (begin
	  (er "get-sensors-name-list" "the argument is not a structure defining sensors")
	  '*error-get-sensors-name-list*)))

;; a function that retrieves an input port for a sensors structure.
;;
(define (get-sensors-input-file sen)
   (if (sensors? sen)
       (sensors-input-file sen)
       (begin
	  (er "get-sensors-input-file" "the argument is not a structure defining sensors")
	  '*error-get-sensors-input-file*)))

;; a function that opens an input file, reads the line with sensor names,
;; and returns a structure for sensors
;;
(define (open-sensors file-name)
   (let ((file (open-input-file file-name)))
      (let ((names (read-name-list file)))
	 (make-sensors file names))))
      
;; a function that closes an input port of the structure for sensors.
;;
(define (close-sensors sensors-struct)
   (let ((port (get-sensors-input-file sensors-struct)))
      (close-input-port port)))

;; a function that returns a list containing a new data set.
;;
(define (get-new-data sensors-struct)
   (read-data-set-list (get-sensors-input-file sensors-struct)
		       (get-sensors-name-list sensors-struct)))

;; A function that returns a list of facts built from data
;; of the next time step.
;;
(define (get-new-facts sensors-struct)
   (let ((data (get-new-data sensors-struct)))
      (let loop ((l data)(res (list)))
	 (if (null? l)
	     res
	     (let ((name (caar l))(value (cadar l)))
		(let ((mat (make-loopy-matrix-with-content (vector (length value))
							   (list->vector value))))
		   (let ((f (make-fact (list name 't) mat)))
		      (loop (cdr l) (append res (list f))))))))))


(define cosmos-data #f)

(define-structure data-control sensors_data
                               given_facts
			       current_facts
			       previously_seen_facts_buffer
			       last_tick)

;; A function that creates a data-control function based on a
;; sensor port and given facts. Note that given facts must be
;; rearranged as a list of time steps, where a time step is a
;; list of facts given at a single time step. The function
;; returns the data-control structure. 
;; 
(define (create-data-control sen facts)

   
   (define (rearrange-facts f)

      (define (get-fact-time elem)
	 (if (fact? elem)
	     (cadr (fact-term elem))
	     (begin
		(er "get-fact-time" "the argument is not a fact")
		'*error-argument*)))
      
      (define (convert-to-number term)
	(cond ((eq? term (string->symbol (number->string 0))) 0)
	      ((and (list? term) (not (null? term)) (eq? (car term) 'succ)) (+ 1 (convert-to-number (cadr term))))
	      (else (er "convert-to-number" "the argument is empty")
		    '*error-argument*)))

      (let loop ((l f)
		 (cur-time (get-fact-time (car f)))
		 (time-step (list 0)) ;; the first element is 0 time away from the previous (there is none)
		 (res '()))
	 (if (null? l)
	     (append res (list time-step))
	     (let* ((elem (car l))
		    (time (get-fact-time elem))
		    (gap (- (convert-to-number time) 
			    (convert-to-number cur-time))))
		(if (zero? gap)
		    (loop (cdr l)
			  cur-time
			  (append time-step (list elem))
			  res)
		    (loop (cdr l)
			  time
			  (list gap elem)
			  (append res (list time-step))))))))
   
   (if (sensors? sen)
       ;; rearrange given facts into a list of lists 
       (let ((given_facts (rearrange-facts facts)))
	  ;; make a data-control structure; note that current_facts is empty for now.
	  (make-data-control sen given_facts '() '() #f))
       (begin
	  (er "create-data-control" "the first argument is not sensors")
	  '*error-argument*)))

;; A function that returns the facts of the current time step.
;;
(define (get-current-facts dc)
   (if (data-control? dc)
       (let ((cur (data-control-current_facts dc)))
	 (if (null? cur)
	     '()
	     (cdr cur)))
       (begin
	  (er "get-current-facts" "the argument is not a data-control structure")
	  '*error-argument*)))

;; A function that returns the facts of the current time step.
;;
(define (get-current-gap dc)
   (if (data-control? dc)
       (let ((cur (data-control-current_facts dc)))
	 (if (null? cur)
	     #f
	     (car cur)))
       (begin
	  (er "get-current-facts" "the argument is not a data-control structure")
	  '*error-argument*)))

;; A function that pushes current facts to the buffer of previously
;; seen facts, gets a set of facts from the next time step of the
;; given facts, assignes that set to current_facts, and returns it.
;;
(define (get-next-given-facts dc)
   (if (data-control? dc)
       (let ((cur (data-control-current_facts dc))
	     (buf (data-control-previously_seen_facts_buffer dc))
	     (giv (data-control-given_facts dc)))
	  (if (not (null? cur))
	      (data-control-previously_seen_facts_buffer-set! dc (append buf (list cur))))
	  (if (null? giv)
	      (begin
		 (data-control-current_facts-set! dc '())
		 '())
	      (let ((new-cur (car giv))(new-giv (cdr giv)))
		 (data-control-current_facts-set! dc new-cur)
		 (data-control-given_facts-set! dc new-giv)
		 ;; the current tick is a time of the first fact of the time step (second element of the list)
		 (data-control-last_tick-set! dc (cadr (fact-term (cadr new-cur))))
		 new-cur)))
       (begin
	  (er "get-next-given-facts" "the argument is not a data-control structure")
	  '*error-argument*)))

;; A function that pushes current facts to the buffer of previously
;; seen facts, gets a set of facts from the next time step, assignes that
;; set to current_facts, and returns it. Note that the function tries
;; to get the new facts from the given facts. If the given facts are
;; empty, it reads new facts from sensor data.
;;
(define (get-next-facts dc)
   (if (data-control? dc)
       (let ((cur (data-control-current_facts dc))
	     (buf (data-control-previously_seen_facts_buffer dc))
	     (giv (data-control-given_facts dc))
	     (sens (data-control-sensors_data dc)))
	  (if (not (null? cur))
	      (data-control-previously_seen_facts_buffer-set! dc (append buf (list cur))))
	  (if (null? giv)
	      (let ((new-facts (get-new-facts sens)))
		 (if (null? new-facts)
		     (begin
			(data-control-current_facts-set! dc '())
			'())
		     (let ((next-data-step (cons 1 new-facts)))
			(data-control-current_facts-set! dc next-data-step)
			next-data-step)))
	      (let ((new-cur (car giv))(new-giv (cdr giv)))
		 (data-control-current_facts-set! dc new-cur)
		 (data-control-given_facts-set! dc new-giv)
		 new-cur)))
       (begin
	  (er "get-next-facts" "the argument is not a data-control structure")
	  '*error-argument*)))

;; A function that flushes the buffer of previously seen facts.
;;
(define (forget-previously-seen-facts dc)
   (if (data-control? dc)
       (begin
	 (data-control-previously_seen_facts_buffer-set! dc '())
	 (data-control-current_facts-set! dc '()))
       (begin
	  (er "forget-previously-seen-facts" "the argument is not a data-control structure")
	  '*error-argument*)))


;; A function that puts a buffer of previously seen facts in front of the
;; list of given facts and makes a set of current facts to be the facts from
;; the first time step of the given facts. The buffer of previously seen facts
;; becomes empty.
;; ????? this means "given_facts" must be reorganized from the facts_list!!!!!!!!!!!
;;
(define (reuse-previously-seen-facts dc) ;;???? do i need this function????
   (if (data-control? dc)
       (let ((seen (data-control-previously_seen_facts_buffer dc))
	     (cur (data-control-current_facts dc))
	     (giv (data-control-given_facts dc)))
	  (data-control-previously_seen_facts_buffer-set! dc '())
	  (data-control-current_facts-set! dc '())
	  (if (null? cur)
	      (data-control-given_facts-set! dc (append seen giv))
	      (data-control-given_facts-set! dc (append seen (list cur) giv))))
       (begin
	  (er "reuse-previously-seen-facts" "the argument is not a data-control structure")
	  '*error-argument*)))

;; the definition of a data window
;;
;(define data-window (list))

;; a function that returns a current size of a data window
;;
;(define (get-data-window-size window)
;   (if (list? window)
;       (length window)
;       (begin
;	  (er "get-data-window-size" "the argument is not a list")
;	  '*error-argument*)))

;; a function that returns a new data window that is a concatenation of
;; the previous window and a new data set.
;;
;(define (add-data-to-window window data)
;   (if (and (list? window) (list? data))
;       (append window (list data))
;       (begin
;	  (er "add-data-to-window" "the argument is not a list")
;	  '*error-argument*)))

;; a function that returns a new data window by removing the first
;; element from the current window.
;;
;(define (remove-data-from-window window)
;   (if (list? window)
;       (cdr window)
;       (begin
;	  (er "remove-data-from-window" "the argument is not a list")
;	  '*error-argument*)))

;(define (fill-up-window data-window sens)
;   (let loop ((window data-window)(i (get-data-window-size data-window)))
;      (if (< i window-size)
;	  ;; if the data window is not full, fill it up
;	  (loop (add-data-to-window window (get-new-data sens))(+ i 1))
;	  window)))









;(let ((sens (open-sensors "data.txt")))
;   (display "The sensors structure: ")
;   (pretty-print sens)
;   (set! data-window (fill-up-window data-window sens))
;   (display "The filled up data window: ")
;   (pretty-print data-window)
;   (set! data-window (remove-data-from-window data-window))
;   (set! data-window (add-data-to-window data-window (get-new-data sens)))
;   (display "The updated data window: ")
;   (pretty-print data-window)
;   (display "The new data set: ")
;   (pretty-print (get-new-data sens))
;   (close-sensors sens)
;   )
