;;;;;
;;;;; The file containing definitions and operations for
;;;;; making an inference across a network
;;;;;


;; max : number, number -> boolean
;; A function that finds a maximum between two numbers;
;; it could change in the future, otherwise it is not needed (same as the original).

(define (max a b)
  (if (> a b)
      a
      b))

;; update-cluster-node-list : list cluster -> float
;; A function that computes the new messages from each cluster to each variable it is connected with,
;; and computes an updated potential for each cluster, and returns the maximal change
;; between the new and the old potentials of each cluster.

(define (update-cluster-node-list clstr_lst)

   ;; update-one-cluster-node : cluster -> float
   ;; A subroutine that computes messages from the cluster to every neighboring variable, and updates
   ;; the potential of the cluster, and returns the difference between the new and the old
   ;; potentials; if the argument is not a cluster node, it returns -1.
   
   (define (update-one-cluster-node clstr)
      (if (is-cluster-node? clstr)
	  ;; remember current potential and cpt (or its approximation from the learnable node) of the cluster
	  (let ((old_p (get-cluster-node-potential clstr))(p (get-cluster-node-table clstr)))
	     (tr "+++++++++++++++++++++++++ updating a cluster with ID " (get-cluster-node-id clstr))
	     (tr "the current cpt is " (get-cluster-node-table clstr))
	     ;; iterate through neighbors of the cluster
	     (let loop ((l (get-cluster-node-neighbors clstr))(i 0))
		(if (not (null? l))
		    ;; for each neighbor, multiply cluster's cpt with the messages from all OTHER neighbors
		    (let ((cpt (get-cluster-node-table clstr)))
		       (do ((vars (get-cluster-node-neighbors clstr) (cdr vars))(j 0 (+ j 1)))
			   ((null? vars))
			   (if (not (= i j))
			       (begin
				  (tr "    multiplying a cpt " cpt)
				  (tr "    with a message " (get-edge-message-from-variable (car vars)))
				  (tr "    along dimension " j)
				  (set! cpt (matrix-times-vector cpt (get-edge-message-from-variable (car vars)) j))
				  (tr "         resulting cpt is " cpt))))
		       ;; marginalize the result and send its normalized version back to the current neighbor
		       (let ((msg (marginalize-matrix cpt i)))
			  (let ((norm-msg (normalize-vector msg)))
			    (if norm-msg
				(set! msg norm-msg)
				(begin
				  (display "WARNING: detected DIVISION BY ZERO: message CLUST->VAR\n")
				  (set! msg (just-above-zero-vector msg)))))
			  (tr "making a new message FROM the cluster TO a variable with ID " (get-variable-node-id (get-edge-variable (car l))))
			  (tr "the original message is " (get-edge-message-from-cluster (car l)))
			  (edge-message-from-cluster-set! (car l) msg)
			  (tr "the new message is " (get-edge-message-from-cluster (car l))))
		       ;; compute the new potential as a product of the cpt with ALL the messages from cluster's neighbors
		       (set! p (matrix-times-vector p (get-edge-message-from-variable (car l)) i))
		       (loop (cdr l) (+ i 1)))))
	     (tr "the original potential of the cluster is " (get-cluster-node-potential clstr))
	     ;; scale the potential
	     (let* ((mat-dim (get-matrix-dimensions p))
		    (mat-data (get-matrix-data p))
		    (norm-data (normalize-vector mat-data)))
	       (if norm-data
		   (cluster-node-potential-set! clstr (make-loopy-matrix-with-content mat-dim norm-data))
		   (begin 
		     (display "WARNING: detected DIVISION BY ZERO: potential CLUST\n")
		     (cluster-node-potential-set! clstr (make-loopy-matrix-with-content mat-dim 
											(just-above-zero-vector mat-data))))))
	     (tr "the new potential of the cluster is " (get-cluster-node-potential clstr))
	     ;; compute and return the (euclidean) distance between the new and the old potential
	     (let ((dist (vector-distance (get-matrix-data old_p) (get-matrix-data (get-cluster-node-potential clstr)))))
		(tr "the change in potential is " dist)
		dist))
	  (begin
	     (er "update-cluster-node-list" "the argument is not a cluster node")
	     -1)))
   ;; the body of update-cluster-node-list function
   (let loop ((l clstr_lst)(acc 0.0))
      (cond
	 ((null? l) acc)
	 ;; iterate through each cluster and return the greatest potential change among them
	 ((is-cluster-node? (car l)) (loop (cdr l) (max acc (update-one-cluster-node (car l)))))
	 (else  acc))))

;; update-variable-node-list : list variable-node -> float
;; A function that makes an update for each element of the list of variable nodes
;; and returns the maximal change

(define (update-variable-node-list var_lst)

   ;; update-one-variable-node : variable-node -> float
   ;; A suroutine that updates a single variable node and returns the change between its old and new beliefs;
   ;; it returns -1, if the argument is not a variable node.
   
   (define (update-one-variable-node var)
      (if (is-variable-node? var)
	  ;; store the current belief and initialize the new belief to a vector of ones
	  (let ((old_b (get-variable-node-belief var))(b (vector-of-ones (get-variable-node-dim var))))
	     (tr "+++++++++++++++++++++++++ updating a variable with ID " (get-variable-node-id var))
	     ;; for every neighbor of the variable node
	     (let loop ((l (get-variable-node-neighbors var))(i 0))
		(if (not (null? l))
		    ;; multiply the messages from all neighbors EXCEPT the current neighbor
		    (let ((msg (vector-of-ones (get-variable-node-dim var))))
		       (do ((cls (get-variable-node-neighbors var) (cdr cls))(j 0 (+ j 1)))
			   ((null? cls))
			   (if (not (= i j))
			       (set! msg (vector-times-vector (get-edge-message-from-cluster (car cls)) msg))))
		       ;; send the normalized result to the current neighbor
		       (let ((norm-msg (normalize-vector msg)))
			 (if norm-msg
			     (set! msg norm-msg)
			     (begin
			       (display "WARNING: detected DIVISION BY ZERO: message VAR->CLUST\n")
			       (set! msg (just-above-zero-vector msg)))))
		       (tr "making a new message FROM the variable TO a cluster with ID " (get-cluster-node-id (get-edge-cluster (car l))))
		       (tr "the original message is " (get-edge-message-from-variable (car l)))
		       (edge-message-from-variable-set! (car l) msg)
		       (tr "the new message is " (get-edge-message-from-variable (car l)))
		       ;; compute the new belief as a product of ALL messages from the neighbors
		       (set! b (vector-times-vector (get-edge-message-from-cluster (car l)) b))
		       (loop (cdr l) (+ i 1)))))
	     ;; normalize the new belief and return its distance to the old belief
	     (let ((norm-b (normalize-vector b)))
	       (if norm-b
		   (set! b norm-b)
		   (begin
		     (display "WARNING: detected DIVISION BY ZERO: belief VAR\n")
		     (set! b (just-above-zero-vector b)))))
	     (tr "the original belief of the variable is " (get-variable-node-belief var))
	     (variable-node-belief-set! var b)
	     (tr "the new belief of the variable is " (get-variable-node-belief var))
	     (let ((dist (vector-distance old_b b)))
		(tr "the change in belief is " dist)
		dist))
	  (begin
	     (er "update-variable-node-list" "the argument is not a variable node")
	     -1)))
   ;; the body of update-variable-node-list function
   (let loop ((l var_lst)(acc 0.0))
      (cond
	 ((null? l) acc)
	 ;; iterate through each variable node and return the greatest belief change among them
	 ((is-variable-node? (car l)) (loop (cdr l) (max acc (update-one-variable-node (car l)))))
	 (else acc))))

;; update-learnable-node-list : list learnable-node -> float
;; A function that makes an update for each element of the list of learnable nodes
;; and returns the maximal table change between them.

(define (update-learnable-node-list learn_lst)

   ;; update-one-learnable-node : learnable-node -> float
   ;; A subroutine that updates a single learnable node and return its table change;
   ;; or returns -1 if the argument is not a learnable node.
   
   (define (update-one-learnable-node learn)
      (if (is-learnable-node? learn)
	  ;; record the current table
	  (let ((old_t (get-learnable-node-table learn))(t (make-loopy-matrix (get-matrix-dimensions (get-learnable-node-table learn)))))
	     (tr "+++++++++++++++++++++++++ updating a learnable with ID " (get-learnable-node-id learn))
	     ;; iterate through the list of neighbors and add their potentials together
	     (let loop ((l (get-learnable-node-neighbors learn)))
		(if (not (null? l))
		    (begin 
		       (set! t (matrix-plus-matrix (get-cluster-node-potential (car l)) t))
		       (loop (cdr l)))))
	     ;; set the result as a new table
	     (tr "the original table (conditionally normalized) of the learnable is " (conditionally-normalize-matrix (get-learnable-node-table learn)))
	     (learnable-node-table-set! learn t) 
	     (tr "the new table (conditionally normalized) of the learnable is " (conditionally-normalize-matrix (get-learnable-node-table learn)))
	     ;; compute and return the distance between the new and the old tables
	     (let ((dist (vector-distance (get-matrix-data old_t) (get-matrix-data t))))
		(tr "the change in table is " dist)
		dist))
	  (begin
	     (er "update-learnable-node-list" "the argument is not a learnable node")
	     -1)))
   ;; the body of update-learnable-node-list function
   (let loop ((l learn_lst)(acc 0.0))
      (cond
	 ((null? l) acc)
	 ;; for each element on the list, update it and get the biggest change
	 ((is-learnable-node? (car l)) (loop (cdr l) (max acc (update-one-learnable-node (car l)))))
	 (else acc))))

;; update-step : network -> float
;; A function that makes one iteration of updating of the clusters, variables, and learnable
;; nodes of the network, and returns the maximal change of all; or returns -1
;; if the argument is not a network.

(define (update-step net) 
  (if (is-network? net)
      (let ((cl (update-cluster-node-list (get-network-clusters net))))
	(let ((va (update-variable-node-list (get-network-variables net))))
	  (let ((le (update-learnable-node-list (get-network-learnable net))))
	    (tr "+++++++++++++++++++++++++ the change in the clusters is " cl)
	    (tr "+++++++++++++++++++++++++ the change in the variables is " va)
	    (tr "+++++++++++++++++++++++++ the change in the learnables is " le)
	    (max cl (max va le))
	    )))
      (begin (er "update-step" "the argument is not a network")
	     -1)))

;; run-update : network, integer, float -> integer
;; A function that iteratively updates the network, and stops
;; the computation if the number of iterations reach the
;; given integer or the change after the iteration is less 
;; than the threshold and return the number of iterations;
;; or spits an error if the arguments are not valid.

(define (run-update net max_iter threshold) 
   (cond
      ((not (is-network? net))(er "run-update" "the first argument is not a network") '*error-run-update*)
      ((not (integer? max_iter))(er "run-update" "the second argument is not an integer") '*error-run-update*)
      ((not (real? threshold))(er "run-update" "the third argument is not a real number") '*error-run-update*)
      (else (let loop ((i 0)(c (update-step net)))
	       (tr "+++++++++++++++++++++++++ +++++++++++++++++++++++++ ITERATION " i)
	       (cond
		  ((= i max_iter) (tr "+++++++++++++++++++++++++ exceeded the maximum of iterations " i) i)
		  ((< c threshold) (tr "+++++++++++++++++++++++++ the threshold is greater than the change " c) i)
		  (else (loop (+ i 1) (update-step net))))))))

;; A function that updates the distributions of a knowledge base with learned distributions
;;
(define (update-knowledge net)
   (if (network? net)
       (let loop ((l (network-learnable net)))
	  (if (not (null? l))
	      (let ((learn (car l)))
		 (rule-learned-set! (learnable-node-loopy_rule learn)
				    (conditionally-normalize-matrix (learnable-node-table learn)))
		 (loop (cdr l)))))
       (begin
	  (er "update-knowledge" "the argument is not a network")
	  '*error-argument*)))
