;;;;;
;;;;; The file containing definitions and operations
;;;;; for making a knowledge base
;;;;;


;; hash : string, integer -> integer
;; given a key, to return a corresponding index of a hash table.
;; This function uses a string hashing function, provided by the
;; Gambit. 

(define (hash key hash-length)
   (modulo (symbol-hash key) hash-length))

;; make-hash-table : integer, hash-function
;; A function that performes one of possible operations
;; on a hash table:
;;    add-entry!
;;    get-first-entry
;;    remove-entry!
;;    display-table
;;    clear-table!

(define make-hash-table
   (lambda (hash-length hash-function)
      (let ((hash-table (make-vector hash-length '())))
	 (lambda (operation . args)
	    (case operation
	       ((add-entry!) (let* ((hash-key (car args))
				    (index (hash-function hash-key hash-length)))
				(vector-set! hash-table
					     index
					     (append (vector-ref hash-table index)
						     (list (cons hash-key (cadr args)))))))
	       ((get-first-entry) (let* ((hash-key (car args))
					 (index (hash-function hash-key hash-length)))
				     (cond ((assoc hash-key (vector-ref hash-table index)) => cdr)
					   (else '*hash-table-search-failed*))))
	       ((get-every-entry) (let* ((hash-key (car args))
					 (index (hash-function hash-key hash-length))
					 (slot (vector-ref hash-table index)))
				     (let loop ((l slot)(acc '()))
					(let ((m (member (assoc hash-key l) slot)))
					   (if (not m)
					       (if (null? acc)
						   '*hash-table-search-failed*
						   acc)
					       (loop (cdr m) (append acc (cons (cdr (car m)) '()))))))))
	       ((get-all-entries) (let loop ((index 0)(res '()))
				     (if (= index hash-length)
					 res
					 (let ((slot (vector-ref hash-table index)))
					    (if (not (null? slot))
						(loop (+ index 1) (append res slot))
						(loop (+ index 1) res))))))
	       ((remove-entry!) (let* ((hash-key (car args))
				       (index (hash-function hash-key hash-length)))
				   (let loop ((slot (vector-ref hash-table index)) (acc '()))
				      (cond ((null? slot) (vector-set! hash-table index (reverse acc)))
					    ((equal? hash-key (caar slot)) (loop (cdr slot) acc))
					    (else (loop (cdr slot) (cons (car slot) acc)))))))
	       ((display-table) (do ((index 0 (+ index 1)))
				    ((= index hash-length))
				    (let ((slot (vector-ref hash-table index)))
				       (if (not (null? slot))
					   (begin
					      (display "Slot #")
					      (display index)
					      (display ": ")
					      (pretty-print slot))))))
	       ((clear-table!) (do ((index 0 (+ index 1)))
				   ((= index hash-length))
				   (vector-set! hash-table index '())))
	       (else (display "ERROR: hash-table: no such operation\n")
		     '*hash-table-error*)
	       )))))


;; a structure definition for the loopy definition

(define-structure definition name dim values)

;; make-loopy-definition : symbol, integer, list symbol -> loopy-definition
;; A function that creates a new definition whose name, dimension, and values are given
;; as inputs;
;; it returns an error if the arguments are invalid.
;;
;; Note that make-definition is provided by Gambit

(define (make-loopy-definition n i lst)
  (pretty-print "Making a loopy definition");;;
   (cond
      ((not (symbol? n))(begin
			    (er "make-loopy-definition" "the first argument is not a symbol")
			    '*error-make-definition*))
      ((not (number? i))(begin
			    (er "make-loopy-definition" "the second argument is not a number")
			    '*error-make-definition*))
      ((null? lst)(begin
		     (er "make-loopy-definition" "the third argument is an empty list")
		     '*error-make-definition*))
      (else (make-definition n i lst))))

;; loopy-definition? : definition -> boolean
;; A function that checks whether or not the argument is a valid definition
;;
;; Note that definition?, definition-name, definition-dim, definition-values are provided by Gambit

(define (is-loopy-definition? def)
   (and (definition? def)
	(symbol? (definition-name def))
	(number? (definition-dim def))
	(not (null? (definition-values def)))))

;; get-loopy-definition-name : loopy-definition -> symbol 
;; A function that returns a name of the definition;
;; or an error otherwise.
;;
;; Note that definition-name is provided by Gambit

(define (get-loopy-definition-name def)
   (if (is-loopy-definition? def)
       (definition-name def)
       (begin
	  (er "get-loopy-definition-name" "the argument is not a definition")
	  '*error-get-definition-name*)))

;; get-loopy-definition-dim : loopy-definition -> integer
;; A function that returns a dimension of the definition;
;; or an error otherwise.
;;
;; Note that definition-dim is provided by Gambit

(define (get-loopy-definition-dim def)
   (if (is-loopy-definition? def)
       (definition-dim def)
       (begin
	  (er "get-loopy-definition-dim" "the argument is not a definition")
	  '*error-get-definition-dim*)))

;; get-loopy-definition-values : loopy-definition -> list symbols
;; A function that returns a list of values of the definition;
;; or an error otherwise.
;;
;; Note that definition-values is provided by Gambit

(define (get-loopy-definition-values def)
   (if (is-loopy-definition? def)
       (definition-values def)
       (begin
	  (er "get-loopy-definition-values" "the argument is not a definition")
	  '*error-get-definition-values*)))

;; a structure definition for the loopy query

(define-structure query term)

;; make-loopy-query : list symbol -> loopy-query
;; A function that creates a new query whose term is given
;; as an input;
;; it returns an error if the argument is invalid.
;;
;; Note that make-query is provided by Gambit

(define (make-loopy-query n)
  (pretty-print "Making a loopy query");;;
   (cond
      ((null? n)(begin
		   (er "make-loopy-query" "the argument is an empty list")
		   '*error-make-query*))
      (else (make-query n))))

;; loopy-query? : query -> boolean
;; A function that checks whether or not the argument is a valid query
;;
;; Note that query?, query-term are provided by Gambit

(define (is-loopy-query? q)
   (and (query? q)
	(not (null? (query-term q)))))

;; get-loopy-query-term : loopy-query -> symbol 
;; A function that returns a term of the query;
;; or an error otherwise.
;;
;; Note that query-term is provided by Gambit

(define (get-loopy-query-term q)
   (if (is-loopy-query? q)
       (query-term q)
       (begin
	  (er "get-loopy-query-term" "the argument is not a query")
	  '*error-get-query-term*)))

;; a structure definition for the loopy fact

(define-structure fact term value)

;; make-loopy-fact : list symbol, matrix -> loopy-fact
;; A function that creates a new fact whose term, and values are given
;; as inputs;
;; it returns an error if the arguments are invalid.
;;
;; Note that make-fact is provided by Gambit

(define (make-loopy-fact t v)
   ;; a subroutine that converts something like "emit(0)=a"
   ;; into "emit(0)=[1.0,0.0,0.0]"
   (define (convert-fact-to-matrix term val)
      (let ((def (definitions-hash 'get-first-entry (car term))))
	 (if (eq? def '*hash-table-search-failed*)
	     '*error-convert*
	     (let loop ((v (get-loopy-definition-values def))(i 0))
		(if (null? v)
		    '*error-convert*
		    (if (eq? (car v) val)
			(let* ((dim (get-loopy-definition-dim def))(vec (make-vector dim 0.0)))
			   (vector-set! vec i 1.0)
			   (make-loopy-matrix-with-content (vector dim) vec))
			(loop (cdr v) (+ i 1))))))))
  (pretty-print "Making a loopy fact");;;
   ;; the body
   (cond
      ((and (not (loopy-matrix? v)) (not (symbol? v)))(begin
						     (er "make-loopy-fact" "the second argument is not a matrix nor a symbol")
						     '*error-make-fact*))
      ((null? t)(begin
		   (er "make-loopy-fact" "the first argument is an empty list")
		   '*error-make-fact*))
      (else (if (symbol? v)
		(let ((mat (convert-fact-to-matrix t v)))
		   (if (eq? mat '*error-convert*)
		       '*error-make-fact*
		       (make-fact t mat)))
		(make-fact t v)))))

;; loopy-fact? : fact -> boolean
;; A function that checks whether or not the argument is a valid fact
;;
;; Note that fact?, fact-term, fact-value are provided by Gambit

(define (is-loopy-fact? f)
   (and (fact? f)
	(not (null? (fact-term f)))
	(loopy-matrix? (fact-value f))))

;; get-loopy-fact-term : loopy-fact -> list symbol 
;; A function that returns a term of the fact;
;; or an error otherwise.
;;
;; Note that fact-term is provided by Gambit

(define (get-loopy-fact-term f)
   (if (is-loopy-fact? f)
       (fact-term f)
       (begin
	  (er "get-loopy-fact-term" "the argument is not a fact")
	  '*error-get-fact-term*)))

;; get-loopy-fact-value : loopy-fact -> vector
;; A function that returns a value of the fact;
;; or an error otherwise.
;;
;; Note that fact-value is provided by Gambit

(define (get-loopy-fact-value f)
   (if (is-loopy-fact? f)
       (fact-value f)
       (begin
	  (er "get-loopy-fact-value" "the argument is not a fact")
	  '*error-get-fact-value*)))

(define (print-fact f)
  (if (is-loopy-fact? f)
      (let ((t (fact-term f))(v (fact-value f)))
	(display "<fact:  term:  ")(pretty-print (cons (car t) (list (collapse-time (cadr t)))))
	(display "        value: ")(print-matrix v)
	(display ">\n"))
      (begin
	(er "print-fact" "the argument is not a fact")
	'*error-print-fact*)))

(define (print-facts fs)
  (let loop ((l fs))
    (if (not (null? l))
	(begin
	  (print-fact (car l))
	  (loop (cdr l))))))

;; a structure definition for the loopy rule

(define-structure rule head body matrix learned failure_manager)

;; make-loopy-rule : term, list terms, matrix/symbol -> loopy-rule
;; A function that creates a new rule whose head, body, and matrix are given
;; as inputs;
;; it returns an error if the arguments are invalid.
;;
;; Note that make-rule is provided by Gambit

(define (make-loopy-rule h b m)
  (display "Making a loopy rule: ");;;
  (pretty-print h)
   (cond
      ((null? h)(begin
		   (er "make-loopy-rule" "the first argument is an empty list")
		   '*error-make-rule*))
      ((null? b)(begin
		   (er "make-loopy-rule" "the second argument is an empty list")
		   '*error-make-rule*))
      ((not (or (loopy-matrix? m) (symbol? m)))(begin
					       (er "make-loopy-rule" "the third argument is not a matrix nor a symbol")
					       '*error-make-rule*))
      (else (make-rule h b m #f #f))))

;; loopy-rule? : rule -> boolean
;; A function that checks whether or not the argument is a valid rule
;;
;; Note that rule?, rule-head, rule-body, rule-matrix are provided by Gambit

(define (is-loopy-rule? r)
   (and (rule? r)
	(not (null? (rule-head r)))
	(not (null? (rule-body r)))
	(or (loopy-matrix? (rule-matrix r))
	    (symbol? (rule-matrix r)))))

;; get-loopy-rule-head : loopy-rule -> list symbols 
;; A function that returns a head of the rule;
;; or an error otherwise.
;;
;; Note that rule-head is provided by Gambit

(define (get-loopy-rule-head r)
   (if (is-loopy-rule? r)
       (rule-head r)
       (begin
	  (er "get-loopy-rule-head" "the argument is not a rule")
	  '*error-get-rule-head*)))

;; get-loopy-rule-body : loopy-rule -> list terms
;; A function that returns a body of the rule;
;; or an error otherwise.
;;
;; Note that rule-body is provided by Gambit

(define (get-loopy-rule-body r)
   (if (is-loopy-rule? r)
       (rule-body r)
       (begin
	  (er "get-loopy-rule-body" "the argument is not a rule")
	  '*error-get-rule-body*)))

;; get-loopy-rule-matrix : loopy-rule -> matrix/symbol
;; A function that returns a matrix/symbol of the rule;
;; or an error otherwise.
;;
;; Note that rule-matrix is provided by Gambit

(define (get-loopy-rule-matrix r)
   (if (is-loopy-rule? r)
       (rule-matrix r)
       (begin
	  (er "get-loopy-rule-matrix" "the argument is not a rule")
	  '*error-get-rule-matrix*)))

;; all the loopy definition are stored in the global hash table
;; of the size 256

(define definitions-hash (make-hash-table 256 hash))

;; all the loopy rules are stored in the global hash table
;; of the size 256

(define rules-hash (make-hash-table 256 hash))

;; all the loopy queries are stored in the global list

(define queries-list '())

;; all the loopy facts are stored in the global list

(define facts-list '())

;; A function that adds a query to the global list

(define (add-query q lst)
   (append lst (list q)))

;; A function that adds a fact to the global list

(define (add-fact f lst)
   (append lst (list f)))

;; check-dimensions? : rule/fact, symbol -> boolean
;; A function that checks whether the matrix of the rule, or the vector of the fact
;; is consistent with the definition

(define (check-dimensions? sentence key)
   (let ((term-names '()))
      (case key
	 ;; a loopy fact
	 ('fact (let ((def (definitions-hash 'get-first-entry (car (get-loopy-fact-term sentence)))))
		   (if (eq? def '*hash-table-search-failed*)
		       #f
		       ;; true, if it sums to one, and the length of the vector defined is equal to the actual length
		       (and (= (get-loopy-definition-dim def) (vector-length (get-matrix-data (get-loopy-fact-value sentence))))
			    (sum-to-one? (get-loopy-fact-value sentence))))))
	 ;; a loopy rule
	 ('rule (let ((def (definitions-hash 'get-first-entry (car (get-loopy-rule-head sentence)))))
		   (if (eq? def '*hash-table-search-failed*)
		       #f
		       (if (symbol? (get-loopy-rule-matrix sentence))
			   ;; true, if it is a learnable distribution
			   #t
			   (let ((dimensions (get-matrix-dimensions (get-loopy-rule-matrix sentence))))
			      ;; false if the first dimension of the matrix defined is not equal to the actual first dimension
			      (if (= (get-loopy-definition-dim def)
				     (vector-ref dimensions 0))
				  (let loop ((l (get-loopy-rule-body sentence))(i 1))
				     ;; recursively check other dimensions and summation to one
				     (if (null? l)
					 (sum-to-one? (get-loopy-rule-matrix sentence))
					 (let ((d (definitions-hash 'get-first-entry (caar l))))
					    (if (eq? d '*hash-table-search-failed*)
						'*error-check-dimensionality*
						(if (= (get-loopy-definition-dim d) (vector-ref dimensions i))
						    (loop (cdr l) (+ i 1))
						    #f)))))
				  #f))))))
	 (else (er "check-dimensions?" "unable to recognize the key")
	       '*error-check-dimensions*))))
