;;;;
;;;; The Loopy Logic language parser definition.
;;;; This file is tightly connected with ll_parse.VER.scm
;;;;


;; We need to use lalr code to generate the parser
(include "lalr.scm")

;; The LALR(1) parser for Loopy Logic language
;;

(define loopy-parser
  (lalr-parser
   
   ;; output a parser, called loopy-parser, in a separate file - ll_lalr_parser.yy.scm,
   (output:    loopy-parser "cosmos_lalr_parser.yy.scm")
   ;; there should be no conflicts
   (expect:    0)
   
   ;; token definitions
   (LPAREN RPAREN LCURLY RCURLY LBRACKET RBRACKET NEWLINE BELONGS COMMA EQ GIVEN QUESTION POUND 
	   ID INT FLOAT VAR MATRIX TRACE SUCCESSOR WINDOW_SIZE
	   (left: PLUS MINUS))
   
   (program  (lines)
	     (newline lines)) ;; getting rid of possible newline symbols in the begining of the file
   
   (lines    (lines sentence) : $2
	     (sentence)       : $1)
	   
   ;; depending on the type of a sentence, it is stored in the global structure
   (sentence     (definition   newline)                 : (definitions-hash 'add-entry! (get-loopy-definition-name $1) $1)
		 (fact   newline)                       : (if (check-dimensions? $1 'fact)
							      (set! facts-list (add-fact $1 facts-list))
							      (parser-exit '*error-inconsistent-fact-dimensions*))
		 (rule   newline)                       : (if (check-dimensions? $1 'rule)
							      (rules-hash 'add-entry! (car (get-loopy-rule-head $1)) $1)
							      (parser-exit '*error-inconsistent-rule-dimensions*))
		 (query   newline)                      : (add-queries-from-list $1)
		 (directive   newline)                  : $1
		 (error  NEWLINE)                       : #f)
   
   ;; a new line is a set of one or more \n
   (newline      (NEWLINE)
		 (newline NEWLINE))

   ;; a definition is ID <- { value-list }
   (definition   (ID BELONGS LCURLY value-list RCURLY)  : (let ((d (make-loopy-definition $1 (length $4) $4)))
							     (if (eq? d '*error-make-definition*)
								 (parser-exit "unable to make a definition")
								 d)))

   ;; a value-list is ID or ID,value-list
   (value-list   (ID)                                   : (list $1)
		 (ID COMMA value-list)                  : (cons $1 $3))

   ;; a fact is term=ID or term=vec
   (fact         (term EQ ID)                           : (let ((f (make-loopy-fact $1 $3)))
							     (if (eq? f '*error-make-fact*)
								 (parser-exit "unable to make a fact")
								 f))
		 (term EQ vec)                          : (let ((f (make-loopy-fact $1 (make-loopy-matrix-with-content (list->vector (car $3))
														    (list->vector (cadr $3))))))
							     (if (eq? f '*error-make-fact*)
								 (parser-exit "unable to make a fact")
								 f)))
   ;; a rule is term|term-list=mat or term|term-list=VAR
   (rule         (term GIVEN term-list EQ mat)          : (let ((r (make-loopy-rule $1 $3 (make-loopy-matrix-with-content (list->vector (car $5))
														       (list->vector (cadr $5))))))
							     (if (eq? r '*error-make-rule*)
								 (parser-exit "unable to make a rule")
								 r))
		 (term GIVEN term-list EQ VAR)          : (let ((r (make-loopy-rule $1 $3 $5)))
							     (if (eq? r '*error-make-rule*)
								 (parser-exit "unable to make a rule")
								 r)))
   ;; a term-list is term or term,term-list
   (term-list    (term)                                 : (list $1)
		 (term COMMA term-list)                 : (cons $1 $3))
   ;; a query is term-list?
   (query        (term-list QUESTION)                   : $1)
   ;; a directive is #command
   (directive    (POUND command)                        : $2)
   ;; a command is TRACE
   (command      (TRACE)        : (begin
				     (display "%%%%% Switching tracing to ON...\n\n")
				     (set! trace_flag #t)
				     $1)
		 (SUCCESSOR ID) : (set! successor $2)
		 (WINDOW_SIZE INT) : (set! window-size $2)) ;; here more loopy logic directives (internal commands) can be added
   ;; a term is ID or ID(args)
   (term         (ID)                                   : (list $1)
		 (ID LPAREN args RPAREN)                : (cons $1 $3))
   ;; args is empty or base arg-rest or term arg-rest
   (args         ()                                     : '()
	         (base arg-rest)                        : (cons $1 $2)
	         (term arg-rest)                        : (cons $1 $2))
   ;; arg-rest is ,base arg-rest or ,term arg-rest or empty
   (arg-rest     (COMMA base arg-rest)                  : (cons $2 $3)
	         (COMMA term arg-rest)                  : (cons $2 $3)
	         ()                                     : '())
   ;; a base is VAR or INT
   (base         (VAR PLUS INT)                         : (let loop ((count $3))
							     (if (zero? count)
								 (list 'var $1)
								 (list 'succ (loop (- count 1)))))
	         (INT PLUS VAR)                         : (let loop ((count $1))
							     (if (zero? count)
								 (list 'var $3)
								 (list 'succ (loop (- count 1)))))
	         (VAR MINUS INT)                        : #f;;;;;;;;;;under construction
	         (INT MINUS VAR)                        : #f;;;;;;;;;;under construction
	         (VAR)                                  : (list 'var $1)
	         (INT)                                  : (let loop ((count $1))
							     (if (zero? count)
								 (string->symbol (number->string count))
								 (list 'succ (loop (- count 1))))))
   ;; a number is INT or FLOAT
   (number       (INT)                                  : $1
		 (FLOAT)                                : $1)
   ;; a number-list is number or number,number-list
   (number-list  (number)                               : (list $1)
		 (number COMMA number-list)             : (cons $1 $3))
   ;; a vec is [number-list]
   (vec          (LBRACKET number-list RBRACKET)        : (list (list (length $2)) $2))
   ;; a mat is vec or [mat,mat-list]
   (mat          (vec)                                  : $1
		 (LBRACKET mat COMMA mat-list RBRACKET) : (let ((m (merge (car (cdr $2)) (car (cdr $4)) (car (car $4)))))
							     (if (eq? m '*error-merge*)
								 (parser-exit '*error-merging-matrix*)
								 (list (append (car $2) (list (+ (car (car $4)) 1)))
								       m))))
   ;; mat-list is mat or mat,mat-list
   (mat-list     (mat)                                  : (list (cons 1 (car $1)) (car (cdr $1)))
		 (mat COMMA mat-list)                   : (let ((m (merge (car (cdr $1)) (car (cdr $3)) (car (car $3)))))
							     (if (eq? m '*error-merge*)
								 (parser-exit '*error-merging-matrix*)
								 (list (cons (+ (car (car $3)) 1) (car $1))
								       m))))
   ))

