;;;;;
;;;;; This is the main file that reads in the knowledge file,
;;;;; constructs the model and performs the inference to
;;;;; answer given queries.
;;;;;

;; get-input-filename : void -> string
;; A function that returns a name of the file to read the information
;; from, depending on whether the file name was supplied
;; as a command line argument or not.

(define (get-input-filename)
  (let ((args (cdr (command-line)))) ;; take the command line arguments
    (if (null? args) 
	"input.txt"   ;; if no arguments were given, return "input.txt"
	(car args)))) ;; otherwise return the first argument 

;; print-answer : integer -> void
;; A function that prints the result of the queries
;; and outputs the amount of iterations
;; it took for the algorithm to finish.

(define (print-answer i1 i2 net)
   (display "%%%%% Printing the result...\n")
   (display "\nAfter ")
   (display (+ i1 1))
   (display " & ")
   (display (+ i2 1))
   (display " iterations\n")
   ;; take a global list of queries
   (let loop ((l queries-list)) 
      (if (not (null? l))
	  ;; take the term of the first query on the list
	  (let* ((qterm (get-loopy-query-term (car l))) 
		 ;; take the variable node corresponding to the query term in the network initialized-network
		 (qvar (get-variable-node qterm net)))
	     ;; if such variable node exists
	     (if qvar 
		 (begin
		    (newline)
		    (pretty-print qterm)
		    ;; take a belief corresponding to the current variable
		    (let out ((val (vector->list (get-variable-node-belief qvar))) 
			      ;; take a list of names of the values of the current variable
			      (vname (get-loopy-definition-values (definitions-hash 'get-first-entry (car qterm))))) 
		       (if (null? val)
			   ;; once all values are printed out, iterate to the next query
			   (loop (cdr l)) 
			   (begin
			      (display "   ")
			      (display (car vname))
			      (display ":  ")
			      (pretty-print (car val))
			      ;; once the first value and its name are printed out, iterate to the next 
			      (out (cdr val) (cdr vname))))))
		 (begin
		    (er "print-answer" "could not find the variable node corresponding to a query")
		    '*error-print-answer*))))))

(define cosmos-manager #f)

(define successor 'succ)

;; the main program

(display "\n***** Welcome to Scheme CoSMoS *****\n\n")
(display "%%%%% Checking for the input file...\n\n")

;; read in the name of the file
(let ((name (get-input-filename)))
   (display "%%%%% Attempting to read ")
   (display name)
   (display " ...\n\n")

   (set! cosmos-manager (create-manager name "data.txt" "net.dot" window-size)))
   
;; parse the file; the global knowledge structures are filled now
(parser (manager-program_file cosmos-manager))
(manager-window_size-set! cosmos-manager window-size)
;; if a directive "#trace" was given, print additional info
(if trace_flag
    (begin
       (display "WINDOW SIZE:\n")
       (pretty-print window-size)
       (display "DEFINITIONS:\n")
       (definitions-hash 'display-table)
       (display "QUERIES:\n")
       (pretty-print queries-list)
       (display "FACTS:\n")
       (print-facts facts-list)
;       (pretty-print facts-list)
       (display "RULES:\n")
       (rules-hash 'display-table)
       (newline)))
(let ((sens (open-sensors "data.txt")))
   (display "%%%%% Reading the data file: data.txt\n")
   (let ((dc (create-data-control sens facts-list)))
      (set! cosmos-data dc);;;!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
      )
   (display "%%%%% Constructing Network... \n\n")
   ;; build a network saved in the network field of cosmos-manager
   (if (eq? (complete-network-from-file cosmos-manager cosmos-data 'given) '*error-argument*) 
       (er "construct-network" "there was an error during the network construction")
       (begin
	  (if trace_flag
	      (begin
		 (display "CONSTRUCTED NETWORK:\n\n")
		 (print-network (manager-network cosmos-manager))))
	  (display "%%%%% Running Inferencing...\n\n")
	  ;; training the network; stop after 100 iterations or when the error is less than E-6
	  (let ((iter (run-update (manager-network cosmos-manager) 99 0.000001)))
	     (if (eq? iter '*error-run-update*)
		 (er "main" "there was an error during inferencing")
		 (begin
		    (if trace_flag
			(begin
			   (display "UPDATED NETWORK:\n\n")
			   (print-network (manager-network cosmos-manager))))
		    (display "CONVERGED IN: ")(pretty-print iter)
		    (print-manager cosmos-manager)
		    (visualize-net (manager-network cosmos-manager) (manager-viz_file cosmos-manager))
		    (update-knowledge (manager-network cosmos-manager))
		    (display "UPDATED RULES:\n")
		    (rules-hash 'display-table)
		    (newline)
		    
		    (display "QUERY NETWORK\n\n")
		    (let ((qnet (make-query-net (data-control-last_tick cosmos-data)
						(manager-network cosmos-manager))))
		       (print-network qnet)
		       (let ((iter2 (run-update qnet 99 0.000001)))
			  (print-answer iter iter2 qnet)))
		    )))
	  (display "DATA: \n")(pretty-print cosmos-data);;;;;;;;;;;
	  (forget-previously-seen-facts cosmos-data);;;;;;;;;;;;;;;;;;;
	  (display "DATA: \n")(pretty-print cosmos-data);;;;;;;;;;;
	  (display "FACTS-LIST: \n")(pretty-print facts-list);;;;;;;;;;;;;
	  (let loop ((l facts-list))
	     (if (null? l)
		 (let ((new-facts (get-next-facts cosmos-data)))
		    (if (null? new-facts)
			(pretty-print "%%%%% The operation of CoSMoS Has finished!!!")
			(begin
			   (display "NEW FACTS: \n")(pretty-print new-facts);;;;;;;;;;;;;;
			   (loop new-facts))))
		 (let ((check (check-for-failure cosmos-data)))
		    (if check
			(pretty-print "FAILURE!!!")
			(pretty-print "SUCCESS!!!"))
		    ;;here I need to call (loop facts-list)!!!!!!!!!!!!!!!!!!!
;		    (loop (list));;;;;;;;;;;;;;;;remove it later
		    )))
	  (close-sensors sens))))

