
;; the definition of a manager structure
;; program_file       -- a name of a loopy logic program file                         (string)
;; data_file          -- a name of a input data file                                  (string)
;; viz_file           -- a name of a visualization output file                        (string)
;; network            -- a current LL network                                         (network)
;; frames             -- a list of time frames                                        (list frame)
;; window_size        -- a size of a window of time frames                            (integer)
;; window_slide_step  -- an amount of time steps before next parameter estimation     (integer)

(define-structure manager program_file
                          data_file
                          viz_file
			  network
			  frames
			  window_size
			  window_slide_step)

;; A function that creates a manager structure
;;
(define (create-manager f1 f2 f3 w)
   (make-manager f1 f2 f3 (create-new-network '() '() '()) '() w 1))

;; A function that adds a frame to a manager at the specified position
;; and returns a new list of frames.
;; Note that the new frame is added to the end of the list.
;;
(define (add-frame-to-manager! man fr pos)
   (let loop ((l (reverse (manager-frames man)))(pos pos)(prefix (list)))
      (if (zero? pos)
	  (let ((new-list (reverse (append prefix (cons fr l)))))
	     (manager-frames-set! man new-list)
	     new-list)
	  (if (null? l)
	      (begin
		 (er "add-frame-to-manager!" "unexpectedly reached the end of the list")
		 '*error-internal*)
	      (loop (cdr l) (- pos 1) (append prefix (list (car l))))))))


(define (greater? t1 t2)
   (let ((symb 't))
      (cond ((eq? t1 symb)  
	     (cond ((eq? t2 symb)  #f)
		   ((and (list? t2) (not (null? t2)) (eq? (car t2) 'succ)) #f)
		   ((and (list? t2) (not (null? t2)) (eq? (car t2) 'pred)) #t)
		   (else (er "greater?" "the second argument is not valid")
			 #f)))
	    ((and (list? t1) (not (null? t1)) (eq? (car t1) 'succ))  
	     (cond ((eq? t2 symb)  #t)
		   ((and (list? t2) (not (null? t2)) (eq? (car t2) 'succ)) (greater? (cadr t1) (cadr t2)))
		   ((and (list? t2) (not (null? t2)) (eq? (car t2) 'pred)) #t)
		   (else (er "greater?" "the second argument is not valid")
			 #f)))
	    ((and (list? t1) (not (null? t1)) (eq? (car t1) 'pred))  
	     (cond ((eq? t2 symb)  #f)
		   ((and (list? t2) (not (null? t2)) (eq? (car t2) 'succ)) #f)
		   ((and (list? t2) (not (null? t2)) (eq? (car t2) 'pred)) (greater? (cadr t1) (cadr t2)))
		   (else (er "greater?" "the second argument is not valid")
			 #f)))
	    (else (er "greater?" "the second argument is not valid")
		  #f))))


(define (get-appropriate-frame node man)
   (let ((tick (variable-node-tick node)))
      (let loop ((l (reverse (manager-frames man)))(pos 0))
	 (if (null? l)
	     (let ((fr (create-new-frame tick)))
		(add-frame-to-manager! man fr pos)
		fr)
	     (let ((fr-tick (frame-time_tick (car l))))
		(cond ((equal? tick fr-tick)
		       (car l))
		      ((greater? tick fr-tick)
		       (let ((fr (create-new-frame tick)))
			  (add-frame-to-manager! man fr pos)
			  fr))
		      (else (loop (cdr l) (+ pos 1)))))))))

(define (inside-time-window? var-tick man)
   (cond ((not (manager? man))
	  (er "inside-time-window?" "the second argument is not a manager")
	  '*error-argument*)
	 (else (let ((last-tick (frame-time_tick (car (manager-frames man)))))
		  (or (equal? var-tick last-tick) (greater? var-tick last-tick))))))



(define (complete-network-from-file man dat tag)
   
   (define (convert-to-number term)
      (cond ((eq? term (string->symbol (number->string 0))) 0)
	    ((and (list? term) (not (null? term)) (eq? (car term) 'succ)) (+ 1 (convert-to-number (cadr term))))
	    (else (er "convert-to-number" "the argument is empty")
		  '*error-argument*)))

   (cond ((not (manager? man))
	  (er "complete-network-from-file" "the first argument is not a manager")
	  '*error-argument*)
	 ((not (data-control? dat)) 
	  (er "complete-network-from-file" "the second argument is not a data control")
	  '*error-argument*)
	 (else
	  (if (< (length (manager-frames man))
		 (manager-window_size man))
	      ;; using only given data
	      (cond ((eq? tag 'given) (if (not (null? (data-control-given_facts dat)))
					  (begin
					     (get-next-given-facts dat)
					     (let ((gap (get-current-gap dat)))
					       (let iter ((i gap))
						 (if (> i 0)
						     (begin
						       (shift-time-frames man)
						       (iter (- i 1)))))
					       (expand-network-one-time-step-from-file man dat)
					       (complete-network-from-file man dat 'given)))))
		    ((eq? tag 'complete) (if (null? (data-control-given_facts dat))
					     (let ((new-facts (get-next-facts dat)))
						(if (not (null? new-facts))
						    (let ((gap (car new-facts)))
						      (let iter ((i gap))
							  (if (> i 0)
							      (begin
								 (shift-time-frames man)
								 (iter (- i 1)))))
						       (expand-network-one-time-step-from-file man dat)
						       (complete-network-from-file man dat 'complete))))
					     (begin
						(get-next-given-facts dat)
						(let ((gap (get-current-gap dat)))
						  (let iter ((i gap))
						    (if (> i 0)
							(begin
							  (shift-time-frames man)
							  (iter (- i 1)))))
						  (expand-network-one-time-step-from-file man dat)
						  (complete-network-from-file man dat 'complete)))))
		    (else (er "complete-network-from-file" "unexpected value of the third argument")
			  '*error-argument*))))))
		    

;; A function that removes a frame corresponding to the last time
;; step in the time window and returns the new frame list.
;; Note that this frame is the first one on the list.

(define (forget-last-frame-in-manager! man)
   #f)

;; A function that makes each time frame of a manager one time step "older"
;; and returns the new frame list

(define (shift-time-frames man)
   (define (get-previous-tick tick)
      (cons 'pred (list tick)))
   
   (if (manager? man)
       (let loop ((l (manager-frames man)))
	  (if (not (null? l))
	      (let* ((fr (car l))(older-tick (get-previous-tick (frame-time_tick fr))))
		 (frame-time_tick-set! fr older-tick)
		 (let iter ((vars (frame-variable_nodes fr)))
		       (if (not (null? vars))
			   (let ((var (car vars)))
			      (variable-node-tick-set! var older-tick)
			      (iter (cdr vars)))))
		 (loop (cdr l)))))
       (begin
	  (er "shift-time-frames" "the argument is not a manager")
	  '*error-argument*)))

;; A function that prints a manager.

(define (print-manager man)
   (if (manager? man)
       (begin
	  (display "$$$$$$$$$$$$$$$$$$$$$$$ MANAGER $$$$$$$$$$$$$$$$$$$$$$$\n\n")
	  (display "Loopy Logic Program File:  ")
	  (display (manager-program_file man))
	  (display "\n")
	  (display "Data File:  ")
	  (display (manager-data_file man))
	  (display "\n")
	  (display "Visualization File:  ")
	  (display (manager-viz_file man))
	  (display "\n")
	  (display "Network: \n")
	  (display "Time Window:  \n")
	  (let loop ((l (manager-frames man)))
	     (if (not (null? l))
		 (begin
		    (print-frame (car l))
		    (loop (cdr l)))))
	  (display "Window size:  ")
	  (display (manager-window_size man))
	  (display "\n")
	  (display "Slide Step:  ")
	  (display (manager-window_slide_step man))
	  (display "\n")
	  (display "$$$$$$$$$$$$$$$$$$$$$$$ $$$$$$$ $$$$$$$$$$$$$$$$$$$$$$$\n\n"))
       (begin
	  (er "print-manager" "the argument is not a manager")
	  '*error-argument*)))

