;;;;;
;;;;; The file with matrix definitions and 
;;;;; operations
;;;;;


;; tr : string, object -> void
;; A function that prints out an object depending on its type.

(define (tr txt obj)
   (define (trace-error o)
      (er "trace" "the second argument is invalid"))
   (cond
      ((not (string? txt))(er "trace" "the first argument is not a string"))
      (else (if trace_flag
		(begin
		   (display "TRACE: ")
		   (display txt)
		   ((cond
		       ((string? obj) display)
		       ((number? obj) display)
		       ((vector? obj) print-vector)
		       ((loopy-matrix? obj) print-matrix)
		       (else trace-error))
		    obj)
		   (newline))))))

;; vector-element-product : vector -> number
;; A function that multiplies all the elements
;; of the input vector;
;; it returns 1 if the length is zero;
;; it returns -1 if the argument is not a vector.
;;
;; Example: vector-element-product(<1,2,3,4>)=1x2x3x4=24
;;
;; Note: vector?, vector-length, vector-ref are built in Scheme functions

(define (vector-element-product vec)
   (if (vector? vec)
       (let ((l (vector-length vec)))
	  (if (= l 0)
	      1
	      (let loop ((l (- l 1))(acc 1))
		 (if (< l 0)
		     acc
		     (loop (- l 1) (* acc (vector-ref vec l)))))))
       (let ((error-not-vector -1))
	  (er "vector-element-product" "the argument is not a vector")
	  error-not-vector)))

;; vector-distance : vector, vector -> integer
;; A function that computes the euclidean distance between two vectors;
;; if either of the arguments is not a vector, it returns -1.
;;
;; Note: vector?, vector-length, vector-ref are built in Scheme functions

(define (vector-distance v1 v2)
   (cond
      ((not (vector? v1)) (begin
			     (er "vector-distance" "the first argument is not a vector")
			     -1))
      ((not (vector? v2)) (begin
			     (er "vector-distance" "the second argument is not a vector")
			     -1))
      (else (let loop ((i 0)(acc 0.0))
	       (if (= i (vector-length v1))
		   acc
		   (let ((dif (- (vector-ref v1 i) (vector-ref v2 i))))
		      (loop (+ i 1) (+ acc (* dif dif)))))))))

;; A structure definition of a matrix 
;; 2x3x4 three-dimensional matrix will be represented as
;; [ <2,3,4>,<[0,0,0],[0,0,1],[0,0,2],[0,0,3],[0,1,0],...> ]

(define-structure matrix dims data)

;; the definition of the empty matrix
;;
;; Note: make-matrix is a Gambit provided function to build a structure
;; make-vector is a Scheme built in function

(define empty-loopy-matrix (make-matrix (vector 0) (make-vector 0)))

;; the definition of the empty vector

(define empty-vector (make-vector 0))

;; make-loopy-matrix : vector -> matrix
;; A function that creates a new matrix whose dimensions are given
;; as an input vector d;
;; it returns an empty matrix if the argument is not a vector.
;;
;; Note: this function wraps around Gambit function make-matrix (in
;; case if we decide to go away from Gambit structures, the function
;; make-matrix will be easy to replace

(define (make-loopy-matrix d)
   (if (vector? d)
       (make-matrix d (make-vector (vector-element-product d)))
       (begin
	  (er "make-loopy-matrix" "the argument is not a vector")
	  empty-loopy-matrix)))

;; make-loopy-matrix-with-content : vector, vector -> matrix
;; A function that creates a new matrix whose dimensions and data are given
;; as input vectors;
;; it returns an empty matrix if the argument are invalid.

(define (make-loopy-matrix-with-content di da)
   (cond
      ((not (vector? di))(begin
			    (er "make-loopy-matrix-with-content" "the first argument is not a vector")
			    empty-loopy-matrix))
      ((not (vector? da))(begin
			    (er "make-loopy-matrix-with-content" "the second argument is not a vector")
			    empty-loopy-matrix))
      (else (make-matrix di da))))

;; vector-of-ones : integer -> vector
;; A function that makes a vector filled with 1's with the length equal to the integer;
;; if the argument is not an integer, it returns an empty vector.

(define (vector-of-ones n)
   (if (integer? n)
       (make-vector n 1)
       (begin
	  (er "vector-of-ones" "the argument is not an integer")
	  empty-vector)))

;; matrix-of-ones : vector -> matrix
;; A function that creates a new matrix filled with 1's whose dimensions are given
;; as an input vector; if the argument is not a vector, it returns
;; an empty matrix.

(define (matrix-of-ones d)
   (if (vector? d)
       (make-loopy-matrix-with-content d (make-vector (vector-element-product d) 1))
       (begin
	  (er "matrix-of-ones" "the argument is not a vector")
	  empty-loopy-matrix)))

;; the definition of a random source that is used to define a function
;; returning a random real number between 0 and 1 depending on a random source.

(define rs (make-random-source))
(define rr (random-source-make-reals rs))

;; matrix-of-random : vector -> matrix
;; A function that makes a new matrix filled with random real numbers between 0 and 1
;; whose dimensions are given as an input vector

(define (matrix-of-random d)
   ;; a subroutine making a list of random elements with the length n
   (define (random-list n)
      (let loop ((n n)(acc '()))
	 (cond ((zero? n) acc)
	       (else (loop (- n 1) (cons (rr) acc))))))
   ;; the body of the matrix-of-random function
   (make-loopy-matrix-with-content d (list->vector (random-list (vector-element-product d)))))

;; get-matrix-dimensions : matrix -> vector
;; A function that returns a vector of dimensions of the matrix;
;; it returns an empty vector, if the argument is not a matrix
;;
;; Note that matrix?, matrix-dims are provided by Gambit

(define (get-matrix-dimensions mat)
   (if (matrix? mat)
       (matrix-dims mat)
       (begin
	  (er "get-matrix-dimensions" "the argument is not a matrix")
	  empty-vector)))

;; get-matrix-data : matrix -> vector
;; A function that returns a vector of data of the matrix;
;; it return an empty vector, if the argument is not a matrix
;;
;; Note that matrix?, matrix-data are provided by Gambit

(define (get-matrix-data mat)
   (if (matrix? mat)
       (matrix-data mat)
       (begin
	  (er "get-matrix-data" "the argument is not a matrix")
	  empty-vector)))

;; matrix-ll? : object -> boolean
;; A function that returns true if the input is a valid matrix,
;; and false otherwise.
;;
;; Note that matrix? is provided by Gambit

(define (loopy-matrix? o)
   (if (matrix? o)
       (let ((dim (get-matrix-dimensions o))(data (get-matrix-data o)))
	  (and (vector? dim)
	       (vector? data)
	       (> (vector-length dim) 0)
	       (> (vector-length data) 0)
	       (= (vector-element-product dim) (vector-length data))))
       #f))

;; ########## Supporting function ##########
;; Used in: matrix-times-vector, cond-normalize, mat-marginalize
;;
;; get-vector-tail : integer, vector -> vector
;; A function that computes a vector which is a tail of the input vector
;; starting from (and including) the i+1 element;
;; to return an empty vector if the first and second arguments
;; are not an integer and a vector correspondingly.
;;
;; Example: get-vector-tail(2,< 1,2,3,4,5 >) = < 4,5 >

(define (get-vector-tail i vec) 
   ;; subroutine that recursively computes the tail of a list
   (define (shift lst n)
      (if (null? lst)
	  lst
	  (if (zero? n)
	      (cdr lst)
	      (shift (cdr lst) (- n 1)))))
   ;; the body of the get-vector-tail function
   (cond
      ((not (integer? i))(begin
			    (er "get-vector-tail" "the first argument is not an integer")
			    empty-vector))
      ((not (vector? vec))(begin
			     (er "get-vector-tail" "the second argument is not a vector")
			     empty-vector))
      (else (list->vector (shift (vector->list vec) i)))))

;; ########## Supporting function ##########
;; Used in: marginalize-matrix
;;
;; get-vector-head : integer, vector -> vector
;; A function that computes the vector which is a head of the input vector
;; ending on (and including) the i element;
;; it returns an empty vector if the first and second arguments
;; are not an integer and a vector correspondingly.
;;
;; Example: get-vector-head(2,< 1,2,3,4,5 >) = < 1,2 >

(define (get-vector-head i vec)
   ;; subroutine that recursively computes the head of the list
   (define (chop lst n)
      (let loop ((l lst)(pos 0)(acc '()))
	 (cond ((null? l) acc)
	       ((= pos n) acc)
	       (else (loop (cdr l) (+ pos 1) (append acc (cons (car l) '())))))))
   ;; the body of the get-vector-head function
   (cond
      ((not (integer? i))(begin
			    (er "get-vector-head" "the first argument is not an integer")
			    empty-vector))
      ((not (vector? vec))(begin
			     (er "get-vector-head" "the second argument is not a vector")
			     empty-vector))
      (else (list->vector (chop (vector->list vec) i)))))

;; matrix-times-vector : matrix, vector, integer -> matrix
;; A function that computes a special product between a matrix and a vector:
;; after the matrix dimension corresponding to the given integer
;; (and a given vector) is identified, each projection of the
;; matrix on an element of the dimension is multiplied on a scalar
;; which is a projection of the vector on the element of the dimension.
;; If the arguments are illegal, the empty matrix is returned.
;;
;; Example: matrix-times-vector( [<2,3,2>,<m111,m112,
;;                                   m121,m122,
;;                                   m131,m132,
;;                                   m211,m212,
;;                                   m221,m222,
;;                                   m231,m232>], <x1,x2,x3>, 1 ) =
;; = [<2,3,2>,<m111x1,m112x1,
;;             m121x2,m122x2,
;;             m131x3,m132x3,
;;             m211x1,m212x1,
;;             m221x2,m222x2,
;;             m231x3,m232x3>]

(define (matrix-times-vector mat vec i)
   (cond
      ((not (loopy-matrix? mat))(begin
				(er "matrix-times-vector" "the first argument is not a matrix")
				empty-loopy-matrix))
      ((not (vector? vec))(begin
			     (er "matrix-times-vector" "the second argument is not a vector")
			     empty-loopy-matrix))
      ((not (integer? i))(begin
			    (er "matrix-times-vector" "the third argument is not an integer")
			    empty-loopy-matrix))
      (else (let* ((dim (get-matrix-dimensions mat))(res (make-vector (vector-element-product dim))))
	       (if (and (>= i 0) (< i (vector-length dim)))
		   (if (= (vector-length vec)(vector-ref dim i))
		       (let ((d (vector-element-product (get-vector-tail i dim)))(len (vector-element-product dim)))
			  (do ((j 0 (+ j 1)))
			      ((= j len))
			      (let ((pos (remainder (quotient j d) (vector-ref dim i))))
				 (vector-set! res j (* (vector-ref (get-matrix-data mat) j) (vector-ref vec pos))))))
		       (er "matrix-times-vector" "incompatible dimmensions"))
		   (er "matrix-times-vector" "the third argument is invalid"))
	       (make-loopy-matrix-with-content dim res)))))


;; vector-times-vector : vector, vector -> vector
;; A function that computes the element by element product of equally sized vectors;
;; if at least one argument is not a vector, it returns an empty vector;
;; if the sizes of the vectors are not equal, it returns a vector of zeros
;; of the length of the first input vector.
;;
;; Example: vector-times-vector(<1,2,3,4>,<5,6,7,8>) = <5,12,21,32>

(define (vector-times-vector in out)
   (cond
      ((not (vector? in))(begin
			    (er "vector-times-vector" "the first argument is not a vector")
			    empty-vector))
      ((not (vector? out))(begin
			     (er "vector-times-vector" "the second argument is not a vector")
			     empty-vector))
      (else (let* ((len (vector-length in))(res (make-vector len)))
	       (if (= len (vector-length out))
		   (do ((i 0 (+ i 1)))
		       ((= i len))
		       (vector-set! res i (* (vector-ref in i) (vector-ref out i))))
		   (er "vector-times-vector" "inconsistent sizes of the vectors"))
	       res))))

;; matrix-plus-matrix : matrix, matrix -> void
;; A function that computes the sum of two matrices;
;; if either of the arguments is not a matrix, it returns an empty matrix.

(define (matrix-plus-matrix m1 m2)
   (cond
      ((not (loopy-matrix? m1))(begin
			       (er "matrix-plus-matrix" "the first argument is not a matrix")
			       empty-loopy-matrix))
      ((not (loopy-matrix? m1))(begin
			       (er "matrix-plus-matrix" "the first argument is not a matrix")
			       empty-loopy-matrix))
      (else (let* ((d1 (get-matrix-data m1))
		   (d2 (get-matrix-data m2))
		   (res (make-vector (vector-length d1))))
	  (do ((i 0 (+ i 1)))
	      ((= i (vector-length d1)))
	      (vector-set! res i (+ (vector-ref d1 i) (vector-ref d2 i))))
	  (make-loopy-matrix-with-content (matrix-dims m1) res)))))

;; conditionally-normalize-matrix : matrix -> void
;; A function taht computes the matrix that is conditionally normalized;
;; if the argument is not a matrix, it returns an empty matrix.

(define (conditionally-normalize-matrix mat)
   (if (loopy-matrix? mat)
       (let* ((s (get-matrix-dimensions mat))
	     (data (matrix-data mat))
	     (l (vector-element-product (get-vector-tail 0 s)))
	     (len (vector-element-product s))
	     (tmp (make-vector len))
	     (dim (vector-ref s 0)))
	  (do ((i 0 (+ i 1)))
	      ((= i len))
	      (let loop ((j 0)(acc 0.0))
		 (if (= j dim)
		     (vector-set! tmp i (/ (vector-ref data i) acc))
		     (loop (+ j 1) (+ acc (vector-ref data (+ (remainder i l) (* j l)))))
		     )))
	  (make-loopy-matrix-with-content s tmp))
       (begin (er "conditionally-normalize-matrix" "the argument is not a matrix")
	      empty-loopy-matrix)))

;; ########## Supporting function ##########
;; Used in: normalize-vector, normalize-matrix
;;
;; vector-element-summation : vector -> integer
;; A function that computes the sum of all elements of the vector;
;; it returns -1 if the argument is not a vector.

(define (vector-element-summation vec)
   (if (vector? vec)
       (let loop ((i (- (vector-length vec) 1))(acc 0))
	  (if (< i 0)
	      acc
	      (loop (- i 1) (+ acc (vector-ref vec i)))))
       (begin (er "vector-element-summation" "the argument is not a vector")
	      -1)))

;; normalize-vector : vector -> vector
;; A function that computes the normalization of a vector;
;; it returns #f if there is division on zero;
;; if the argument is not a vector, it returns
;; an empty vector.

(define (normalize-vector vec)
   (if (vector? vec)
       (let ((s (vector-element-summation vec)))
	 (cond ((zero? s) #f)
	       ((= s 1) vec)
	       (else (let ((s (/ 1.0 s))
			   (len (vector-length vec))
			   (res (make-vector (vector-length vec))))
		       (do ((i 0 (+ i 1)))
			   ((= i len))
			 (vector-set! res i (* (vector-ref vec i) s)))
		       res))))
       (begin (er "normalize-vector" "the argument is not a vector")
	      empty-vector)))

;; A function that returns a vector filled with a very small number
;;

(define (just-above-zero-vector vec)
  (if (vector? vec)
      (let ((len (vector-length vec))
	    (val 0.0000000001))
	(make-vector len val))
      (begin (er "just-above-zero-vector" "the argument is not a vector")
	     empty-vector)))

;; normalize-matrix : matrix -> matrix
;; A function that computes the normalization of a matrix;
;; if the argument is not a matrix, it returns an empty matrix.

(define (normalize-matrix mat)
   (if (loopy-matrix? mat)
       (let* ((vec (matrix-data mat))
	      (res (make-vector (vector-length (matrix-data mat))))
	      (s (vector-element-summation vec)))
	  (if (not (= s 1))
	      (let ((s (/ 1.0 s))(len (vector-length vec)))
		 (do ((i 0 (+ i 1)))
		     ((= i len))
		     (vector-set! res i (* (vector-ref vec i) s)))
		 (make-loopy-matrix-with-content (matrix-dims mat) res))
	      mat))
       (begin (er "normalize-matrix" "the argument is not a matrix")
	      empty-loopy-matrix)))

;; marginalize-matrix : matrix, integer -> vector
;; A function that computes the marginalization of the matrix along the specified
;; dimension; if either of the arguments is invalid, it returns an
;; empty vector.
;;
;; Example: marginalize-matrix( [<2,3,2>,<m111,m112,
;;                                     m121,m122,
;;                                     m131,m132,
;;                                     m211,m212,
;;                                     m221,m222,
;;                                     m231,m232>], 1 ) =
;; = <m111+m112+m211+m212,
;;    m121+m122+m221+m222,
;;    m131+m132+m231+m232>

(define (marginalize-matrix mat i)
  (cond
     ((not (loopy-matrix? mat)) (begin
				(er "marginalize-matrix" "the first argument is not a matrix")
				empty-vector))
     ((not (integer? i)) (begin
			    (er "marginalize-matrix" "the second argument is not an integer")
			    empty-vector))
     (else (let* ((d (vector-ref (get-matrix-dimensions mat) i))
		 (dh (vector-element-product (get-vector-head i (get-matrix-dimensions mat))))
		 (dt (vector-element-product (get-vector-tail i (get-matrix-dimensions mat))))
		 (m (get-matrix-data mat))
		 (res (make-vector d)))
	      (do ((i 0 (+ i 1)))
		  ((= i d))
		  (do ((j 0 (+ j 1)))
		      ((= j dh))
		      (do ((k (* (+ i (* j d)) dt) (+ k 1)))
			  ((= k (* (+ (+ i (* j d)) 1) dt)))
			  (vector-set! res i (+ (vector-ref res i) (vector-ref m k))))))
	      res))))

;; sum-to-one? : mat -> boolean
;; A function that checks whether a given matrix is in fact a probability
;; distribution (sums to one).

(define (sum-to-one? mat)
   #t
;   (cond
;     ((not (loopy-matrix? mat)) (begin
;				(er "sum-to-one?" "the argument is not a matrix")
;				#f))
;     (else (let* ((vdim (get-matrix-dimensions mat))
;		  (d (vector-ref vdim 0))
;		  (drest (vector-element-product (list->vector (cdr (vector->list vdim)))))
;		  (m (get-matrix-data mat)))
;	      (display "VDIM ")
;	      (pretty-print vdim)
;	      (display "D ")
;	      (pretty-print d)
;	      (display "DREST ")
;	      (pretty-print drest)
;	      (display "M ")
;	      (pretty-print m)
;	      (let loop1 ((i 0))
;		 (display "i ")
;		 (pretty-print i)
;		 (if (= i drest)
;		     #t
;		     (let loop2 ((acc 0)(j 0))
;			(display "acc ")
;			(pretty-print acc)
;			(display "j ")
;			(pretty-print j)
;			(display "add ")
;			(if (not (= j d)) (pretty-print (vector-ref m (+ i (* j drest)))))
;			(if (= j d)
;			    (if (= acc 1)
;				(loop1 (+ i 1))
;				#f)
;			    (loop2 (+ acc (vector-ref m (+ i (* j drest)))) (+ j 1)))))))))
   )


;;; ################# Printing Routines ##############

;; print-vector : vector -> void
;; A function that prints out the vector

(define (print-vector vec)
   (if (vector? vec)
       (let ((n (vector-length vec)))
	  (display "< ")
	  (do ((i 0 (+ i 1)))
	      ((= i n))
	      (display (vector-ref vec i))
	      (if (< i (- n 1))
		  (display ", ")))
	  (display " >"))
       (er "print-vector" "the argument is not a vector")))

;; print-matrix : matrix -> void
;; A function that prints out the matrix

(define (print-matrix mat)
   (if (loopy-matrix? mat)
       (begin
	  (display "[")
	  (print-vector (get-matrix-dimensions mat))
	  (display ", ")
	  (print-vector (get-matrix-data mat))
	  (display "]"))
       (er "print-matrix" "the argument is not a matrix")))

;(pretty-print (sum-to-one? (make-loopy-matrix-with-content (vector 3) (vector .2 .33 .47))))
;(pretty-print (sum-to-one? (make-loopy-matrix-with-content (vector 3 2) (vector .2 .1 .33 .5 .47 .4))))
;(pretty-print (sum-to-one? (make-loopy-matrix-with-content (vector 3 2 4) (vector .2 .11 .7 .5 .23 .45 .2 .1 .33 .33 .1 .5 .54 .25 .2 .5 .47 .56 .2 0 .23 .3 .6 .4))))
;(pretty-print (sum-to-one? (make-loopy-matrix-with-content (vector 3 2) (vector .2 .8 .7 .1 .1 .1))))
