;;;;;
;;;;; The file containing routines for
;;;;; constructing a network
;;;;;


;; A function that checks if a term is a variable
;;
(define (is-variable? term)
   (and (list? term)
	(= (length term) 2)
	(eq? (car term) 'var)))



;; A function that checks if a term is a constant
;;
(define (is-constant? term)
   (symbol? term))



;; A function that checks if a term is a successor
;;
(define (is-successor? term)
   (and (list? term)
	(>= (length term) 2)
	(eq? (car term) 'succ)))



;; A function that returns a binding associated with a variable
;;
(define (get-binding var subs)
   (assoc var subs))



;; A function that unifies two terms and return a post general unifier that is a
;; list of pairs of variables and their substitutes.
;;
(define (unify term1 term2)
   (define (unify-acc term1 term2 substitutions)
      (define (plus-minus term)
	 (cons 'succ (list (cons 'pred (list term)))))
      
      (define (occurs? var term)
	 (cond ((equal? var term) #t)
	       ((or (is-variable? term) (is-constant? term) (null? term)) #f)
	       (else (or (occurs? var (car term)) (occurs? var (cdr term))))))
      
      (define (add-substitution var term subs)
	 (cons (cons var term) subs))
      
      (define (match-variable var term subs)
	 (if (equal? var term)
	     subs
	     (let ((bind (get-binding var subs)))
		(cond
		   (bind (unify-acc (cdr bind) term subs))
		   ((occurs? var term) '*failure*)
		   (else (add-substitution var term subs))))))
      ;; the body
      (cond
	 ((eq? substitutions '*failure*) '*failure*)
	 ((null? term1) (if (null? term2)
			    substitutions
			    '*failure*))
	 ((null? term2) '*failure*)
	 ((is-variable? term1)(match-variable term1 term2 substitutions))
	 ((is-variable? term2)(match-variable term2 term1 substitutions))
	 ((is-successor? term1)(if (is-successor? term2)
				   (unify-acc (cdr term1) (cdr term2) substitutions)
				   (if (eq? (string->symbol (number->string 0)) term2)
				       '*failure*
				       (unify-acc term1 (plus-minus term2) substitutions))))
	 ((is-successor? term2)(if (eq? (string->symbol (number->string 0)) term1)
				   '*failure*
				   (unify-acc (plus-minus term1) term2 substitutions)))
	 ((is-constant? term1)(if (equal? term1 term2)
				  substitutions
				  '*failure*))
	 ((is-constant? term2) '*failure*)
	 (else (unify-acc (cdr term1)
			  (cdr term2)
			  (unify-acc (car term1) (car term2) substitutions)))))
   (unify-acc term1 term2 '()))




;; A function that applies the given substitutions on the term.
;;
(define (substitute subs term)
   (define (sub-variable var subs)
      (let ((bind (get-binding var subs)))
	 (if bind
	     (cdr bind)
	     var)))
   ;; the body
   (cond ((null? term) term)
	 ((is-variable? term) (sub-variable term subs))
	 ((is-constant? term) term)
	 (else (cons (substitute subs (car term))
		     (substitute subs (cdr term))))))



;; A function that applies the substitutions to each term in the list.
;;
(define (substitute-list subs lst)
   (let ((f (lambda (l)
	       (substitute subs l))))
      (map f lst)))



;; A function that constructs a subnetwork for a rule and returns
;; a list of new variable nodes reached by the rule.
;;
(define (construct-subnet-from-rule node rule man)
   ;; A supporting function that takes a list of variable nodes and
   ;; constructs a cluster node connecting those; if the matrix is a symbol
   ;; (learnable) it searches for a learnable node (using learnable-mark).
   ;;
   (define (connect-variables-to-clusters rule neighbors net)
      (let ((mat (get-loopy-rule-matrix rule)))
	 (if (symbol? mat)
	     ;; if the matrix is a learnable distribution
	     ;; create a new cluster node
	     (let ((clust (create-new-cluster-node-learnable neighbors)))
		;; add the cluster node to the network
		(add-node-to-network net clust)
		;; try to find a corresponding learnable node
		(let ((learn (get-learnable-node rule net)))
		   ;; there is such a learnable node, add the cluster node to it
		   (if learn (add-cluster-node-to-learnable-node clust learn)
		       ;; otherwise create a new learnable node
		       (let ((new-learn (create-new-learnable-node rule
								   (get-matrix-dimensions (cluster-node-table clust)))))
			  (add-node-to-network net new-learn)
			  (add-cluster-node-to-learnable-node clust new-learn))))
		new-vars)
	     ;; if the matrix is known, create a new cluster node and add it to the network
	     (let ((clust (create-new-cluster-node neighbors mat)))
		(add-node-to-network net clust)))))

   (cond ((not (is-variable-node? node)) '*error-construct-network*)
	 ((not (is-loopy-rule? rule)) '*error-construct-network*)
	 (else
	  ;; get the term of the variable node and try to unify it with the head of the rule
	  (let* ((term (get-variable-node-term node))(subs (unify term (get-loopy-rule-head rule))))
	     (if (eq? subs '*failure*)
		 (list)
		 ;; iterate through the terms of the body of the rule after substitution
		 (let loop ((body (substitute-list subs (get-loopy-rule-body rule)))(neighbors '())(new-vars '()))
		    (if (null? body)
			;; if there are no more terms in the body left, connect the variable nodes with a cluster node
			(begin
			   (connect-variables-to-clusters rule
							  (cons node neighbors)
							  (manager-network man))
			   ;; return newly created variable nodes
			   new-vars)
			;; try to get a variable node corresponding to the term from the body
			(let ((body-term (car body)))
			   (let ((body-tick (get-time-from-term body-term)))
			      ;; check whether the variable node is outside of the time window or not
			      (if (inside-time-window? body-tick man)
				  (let ((body-var (get-variable-node body-term (manager-network man))))
				     ;; if there is such a variable node, add it the list of neighbors and iterate
				     (if body-var
					 (loop (cdr body) (append neighbors (list body-var)) new-vars)
					 ;; otherwise create a new variable node, add it to the list of neighbors, and iterate
					 (let ((nvar (create-new-variable-node body-term
									       (get-loopy-definition-dim (definitions-hash 'get-first-entry (car body-term))))))
					    (add-node-to-network (manager-network man) nvar)
					    ;; add the new variable node to an appropriate frame
					    (add-node-to-frame man nvar)
					    (loop (cdr body) (append neighbors (list nvar)) (append new-vars (list nvar))))))
				  (list)))))))))))




;; A function that merges the second list into the first one: adds every element of the second list
;; into the first list, if there is no such element in the first list.
;;
(define (merge-lists lst1 lst2)
   (define (mem? x ls)
      (cond
	 ((null? ls) #f)
	 ((equal? (get-variable-node-term (car ls)) (get-variable-node-term x)) #t)
	 (else (mem? x (cdr ls)))))
   ;; the body
;   (display "MERGE-LISTS\n");;
;   (display "LIST LST1:")(pretty-print (map variable-node-term lst1));;;;
;   (display "LIST LST2:")(pretty-print (map variable-node-term lst2));;;;
   (if (null? lst1)
       lst2
       (let loop ((l lst2)(acc '()))
;	 (display "LIST L:")(pretty-print (map variable-node-term l));;;;
;	 (display "LIST ACC:")(pretty-print (map variable-node-term acc));;;;
	  (if (null? l)
	      (append lst1 acc)
	      (if (mem? (car l) lst1)
		  (loop (cdr l) acc)
		  (loop (cdr l) (append acc (list (car l))))
		  )))))




;; a function takes a variable node from the list
;; unifies it with rules, builds the network for the rule,
;; and puts the variable nodes of the body on the list;
;; the process iterates until the list is empty.
;;
(define (complete-net-from-vars lst man)
   (let loop ((current lst))
      (if (not (null? current))
	  ;; take the first variable node on the list
	  (let ((var-node (car current)))
	     ;; check if this is a variable node
	     (if (not (is-variable-node? var-node))
		 (begin
		    (er "complete-net-from-vars" "the argument is not a list of variable nodes")
		    '*error-construct-network*)
		 ;; iterate through all the rules
		 (let iter-rules ((r (rules-hash 'get-every-entry (car (get-variable-node-term var-node))))
				  (extended-current (cdr current)))
		    (if (or (null? r) (eq? r '*hash-table-search-failed*))
			;; once we checked every rule, take another variable node from the extended list
			(loop extended-current)
			;; for a rule on the list, construct a subnetwork
			(let ((res (construct-subnet-from-rule var-node (car r) man)))
			   (if (eq? res '*error-construct-network*)
			       '*error-construct-network*
			       ;; add new variable nodes on the list without repetition
			       (iter-rules (cdr r) (merge-lists extended-current res)))))))))))



;; a function that for a given fact creates a subnetwork and calls a function
;; that then completes the network for a given variable node
;;
(define (construct-network-from-fact f man)
   ;; try to find a variable node corresponding to a term
   (let* ((fterm (get-loopy-fact-term f))(fvar (get-variable-node fterm (manager-network man))))
      (if fvar
	  ;; if such a variable node exists, simply create a leaf cluster node
	  (let ((c (create-new-cluster-node (list fvar) (get-loopy-fact-value f))))
	     (add-node-to-network (manager-network man) c)
	     ;; add the fact to an appropriate frame
	     (add-fact-to-frame (get-appropriate-frame fvar man) f))
	  ;; otherwise create a new variable node
	  (let ((new-var (create-new-variable-node fterm
						   (get-loopy-definition-dim (definitions-hash 'get-first-entry (car fterm))))))
	     (add-node-to-network (manager-network man) new-var)
	     ;; add the node to an appropriate frame and get the frame
	     (let ((fr (add-node-to-frame man new-var)))
		;; add the fact to the frame
		(add-fact-to-frame fr f))
	     (add-node-to-network (manager-network man) (create-new-cluster-node (list new-var) (get-loopy-fact-value f)))
	     ;; complete the network for the new variable node
	     (complete-net-from-vars (list new-var) man)))))




;; a function that for a given query creates a subnetwork and calls a function
;; that then completes the network for a given variable node
;;
(define (construct-network-from-query q man)
   ;; try to find a variable node correcponding to a term
   (let* ((qterm (get-loopy-query-term q))(qvar (get-variable-node qterm (manager-network man))))
      (if (not qvar)
	  ;; if no such variable node is found, create a new one
	  (let ((new-var (create-new-variable-node qterm
						   (get-loopy-definition-dim (definitions-hash 'get-first-entry (car qterm))))))
	     (add-node-to-network (manager-network man) new-var)
	     (add-node-to-frame man new-var)
	     ;; complete the network for the new variable node
	     (complete-net-from-vars (list new-var) man)))))




;; A function that expands a network using facts and queries corresponding
;; to the current time step.
;;
(define (expand-network-one-time-step-from-file man dat)
   (define (get-next-tick tick)
      (cons 'succ (list tick)))

   (define (time-of-fact? f t)
      (cond ((not (fact? f))
	     (er "time-of-fact?" "the first argument is not a fact")
	     #f)
	    ((null? t)
	     (er "time-of-fact?" "the second argument is not a time tick")
	     #f)
	    (else (equal? (cadr (fact-term f)) t))))
   
   (define (make-fact-time-relative f)
      (let ((term (fact-term f)))
	 (fact-term-set! f (list (car term) 't))))

   (cond ((not (manager? man))
	  (er "expand-network-one-time-step-from-file" "the first argument is not a manager")
	  '*error-argument*)
	 ((not (data-control? dat)) 
	  (er "expand-network-one-time-step-from-file" "the second argument is not a data control")
	  '*error-argument*)
	 (else
	  (let loop ((facts (get-current-facts dat)))
	     (if (not (null? facts))
		 (let ((f (car facts)))
		    (make-fact-time-relative f)
		    (construct-network-from-fact f man)
		    (loop (cdr facts))))))))

;; A function that expands a network using facts and queries corresponding
;; to the current time step.
;;
(define (expand-network-one-time-step-from-queries man queries)
   (define (make-query-time-relative q)
      (let ((term (query-term q)))
	 (query-term-set! q (list (car term) 't))))

   (cond ((not (manager? man))
	  (er "expand-network-one-time-step-from-file" "the first argument is not a manager")
	  '*error-argument*)
	 (else
	  (let loop ((l queries))
	     (if (not (null? l))
		 (let ((q (car l)))
		    (make-query-time-relative q)
		    (construct-network-from-query q man)
		    (loop (cdr l)))
		 )))))
   
