;;;;;
;;;;; The file with network definitions and
;;;;; operations.
;;;;;


;; make-id : string, integer -> string
;; A function that creates a string from a string and a number by concatenating them together,
;; it is used to create an ID for an object;
;; it returns an empty string if the arguments are invalid.

(define (make-id tag num)
   (cond
      ((not (string? tag))(begin
			     (er "make-id" "the first argument is not a sring")
			     ""))
      ((not (integer? num))(begin
			     (er "make-id" "the second argument is not an integer")
			     ""))
      (else (string-append tag (number->string num)))))

;; the definitions of a variabe node structure
;; id        -- an ID of the variable node                 (string)
;; term      -- a corresponding term                       (list symbols)
;; tick      -- a time tick                                (list symbols)
;; dim       -- an arity of the term                       (integer)
;; belief    -- a belief of the variable node              (vector)
;; neighbors -- a list of the edges to the neighbor nodes  (list edge)
;; host      -- an original variable node that was copied  (variable-node)

(define-structure variable-node id term tick dim belief neighbors host)

;; the definition of the global counter of all variable nodes

(define variable-node-count 0)

;; the definition of the empty variable node
;;
;; Note that make-variable-node is provided by Gambit

(define empty-variable-node (make-variable-node "" (list) (list) 0 empty-vector (list) #f))

;; initialize-belief : int -> vector
;; A function that makes an initial belief vector for a variable node
;; (initially the belief is set to a vector of ones);
;; it returns an empty vector if the argument is not a positive integer.

(define (initialize-belief n)
   (if (integer? n)
       (if (> n 0)
	   (vector-of-ones n)
	   (begin
	      (er "initialize-belief" "the argument is not a positive integer")
	      empty-vector))
       (begin
	  (er "initialize-belief" "the argument is not an integer")
	  empty-vector)))

(define (get-time-from-term term)
   (let ((tick (cdr term)))
;      (if (list? (car tick)) (car tick) tick)))
      (car tick)))

;; create-new-variable-node : list symbols, integer -> variable-node
;; A function that creates a new variable node;
;; it returns an empty variable node if the arguments are not valid.

(define (create-new-variable-node trm dim)
   (cond
      ((null? trm)(begin
		     (er "create-new-variable-node" "the first argument is an empty list")
		     empty-variable-node))
      ((not (integer? dim))(begin
			      (er "create-new-variable-node" "the second argument is not an integer")
			      empty-variable-node))
      (else (let ((n variable-node-count))
	       (set! variable-node-count (+ variable-node-count 1))
	       (let ((tick (get-time-from-term trm)))
		  (make-variable-node (make-id "V" n) trm tick dim (initialize-belief dim) '() #f))))))

;; is-variable-node? : object -> boolean
;; A function that returns true if the object is a valid variable node;
;; it returns false if the object is not valid.
;;
;; Note that variable-node? is provided by Gambit

(define (is-variable-node? var)
   (if (variable-node? var)
       (if (and (integer? (variable-node-dim var))
		(not (negative? (variable-node-dim var))))
	   (and (string? (variable-node-id var))
		(not (null? (variable-node-term var)))
		(vector? (variable-node-belief var)))
	   #f)
       #f))

;; get-variable-node-id : variable-node -> string
;; A function that retrieves the id of a variable node;
;; or it returns an empty string if the argument is not a variable node.
;;
;; Note that variable-node-id is provided by Gambit

(define (get-variable-node-id var)
   (if (is-variable-node? var)
       (variable-node-id var)
       (begin
	  (er "get-variable-node-id" "the argument is not a variable")
	  "")))

;; get-variable-node-term : variable-node -> list symbols
;; A function that retrieves the term of a variable node;
;; or it returns an empty list if the argument is not a variable node.
;;
;; Note that variable-node-term is provided by Gambit

(define (get-variable-node-term var)
   (if (is-variable-node? var)
       (let ((term (variable-node-term var))(tick (variable-node-tick var)))
;	  (if (null? (cdr tick))
;	      (cons (car term) tick)
;	      (cons (car term) (list tick))))
	  (cons (car term) (list tick)))
       (begin
	  (er "get-variable-node-term" "the argument is not a variable")
	  (list))))

;; get-variable-node-dim : variable-node -> integer
;; A function that retrieves the dimensionality of the term of a variable node;
;; or it returns -1 if the argument is not a variable node.
;;
;; Note that variable-node-dim is provided by Gambit

(define (get-variable-node-dim var)
   (if (is-variable-node? var)
       (variable-node-dim var)
       (begin
	  (er "get-variable-node-dim" "the argument is not a variable")
	  -1)))

;; get-variable-node-belief : variable-node -> vector
;; A function that retrieves the belief of a variable node;
;; or it returns an empty vector if the argument is not a variable node.
;;
;; Note that variable-node-belief is provided by Gambit

(define (get-variable-node-belief var)
   (if (is-variable-node? var)
       (variable-node-belief var)
       (begin
	  (er "get-variable-node-belief" "the argument is not a variable")
	  empty-vector)))

;; get-variable-node-neighbors : variable-node -> list CND
;; A function that retrieves the list of neighbors of a variable node;
;; or it returns an empty list if the argument is not a variable node.
;;
;; Note that variable-node-neighbors is provided by Gambit

(define (get-variable-node-neighbors var)
   (if (is-variable-node? var)
       (variable-node-neighbors var)
       (begin
	  (er "get-variable-node-neighbors" "the argument is not a variable")
	  (list))))

;; definitions for a cluster node structure
;; id        -- an ID of a cluster node                           (string)
;; table     -- a conditional probability table of a cluster node (loopy-matrix)
;; potential -- a local potential of a cluster node               (loopy-matrix)
;; neighbors -- a list of edges to the neighbors                  (list edge)
;; learnable -- a list of learnable nodes                         (list learnable-node) (either null or a one-element list)
;; host      -- an original cluster node that was copied          (cluster-node)

(define-structure cluster-node id table potential neighbors learnable host)

;; the definition of the global counter of all cluster nodes

(define cluster-node-count 0)

;; the definitions of the empty cluster node
;;
;; Note that make-cluster-node is provided by Gambit

(define empty-cluster-node (make-cluster-node "" empty-loopy-matrix empty-loopy-matrix (list) (list) #f))

;; initialize-potential : vector -> matrix
;; A function that makes an initial potential matrix for a cluster node
;; (initially the potential is set to a matrix of random reals between zero and one);
;; it returns an empty matrix if the argument is not a vector.

(define (initialize-potential d)
   (if (vector? d)
       (matrix-of-random d)
       (begin
	  (er "initialize-potential" "the argument is not a vector")
	  empty-loopy-matrix)))

;; is-cluster-node? : object -> boolean
;; to return true if an object is a valid cluster node;
;; to return false if the object is not.
;;
;; Note that cluster-node? is provided by Gambit

(define (is-cluster-node? o)
   (if (cluster-node? o)
       (and (string? (cluster-node-id o))
	    (loopy-matrix? (cluster-node-table o))
	    (loopy-matrix? (cluster-node-potential o)))
       #f))

;; definitions for a learnable node structure
;; id         -- an ID of a link                                            (string)
;; loopy_rule -- a rule associated with a learnable node                    (rule)
;; table      -- a current approximation of a conditional probability table (loopy-matrix)
;; neighbors  -- a list of neighbors of a learnable node                    (list cluster-node)

(define-structure learnable-node id loopy_rule table neighbors)

;; the definition of the global counter of all learnable nodes

(define learnable-node-count 0)

;; the definitions of the empty learnable node
;;
;; Note that make-learnable-node is provided by Gambit

(define empty-learnable-node (make-learnable-node "" #f empty-loopy-matrix (list)))

;; init-table : vector -> matrix
;; A function that makes an initial approximation of conditional probability table
;; (it is set to zeros, but it will be changed later, as cluster nodes
;; are added to the network); it returns an empty matrix if the argument is
;; not a vector.

(define (initialize-table d)
   (if (vector? d)
       (make-loopy-matrix d)
       (begin
	  (er "initialize-table" "the argument is not a vector")
	  empty-loopy-matrix)))

;; create-new-learnable-node : list symbols, vector -> learnable-node
;; A function that makes a new learnable node;
;; it returns an empty learnable node if the arguments are not valid.
;;
;; Note that make-learnable-node is provided by Gambit

(define (create-new-learnable-node loopy-rule d)
   (cond
      ((not (rule? loopy-rule))(begin
				  (er "create-new-learnable-node" "the first argument is not a rule")
				  empty-learnable-node))
      ((not (vector? d))(begin
			   (er "create-new-learnable-node" "the second argument is not a vector")
			   empty-learnable-node))
      (else (let ((n learnable-node-count))
	       (set! learnable-node-count (+ learnable-node-count 1))
	       (make-learnable-node (make-id "L" n) loopy-rule (initialize-table d) '())))))

;; is-learnable-node? : object -> boolean
;; A function that returns true if the object is a valid learnable node;
;; it returns false if it is not.
;;
;; Note that learnable-node?, learnable-node-id, learnable-node-rule, learnable-node-table are provided by Gambit

(define (is-learnable-node? o)
   (if (learnable-node? o)
       (and (string? (learnable-node-id o))
	    (rule? (learnable-node-loopy_rule o))
	    (loopy-matrix? (learnable-node-table o)))
       #f))

;; get-learnable-node-id : learnable-node -> string
;; A function that obtains an ID of a learnable node;
;; it returns empty string if the argument is not valid.
;;
;; Note that learnable-node-id is provided by Gambit

(define (get-learnable-node-id learn)
   (if (is-learnable-node? learn)
       (learnable-node-id learn)
       (begin
	  (er "get-learnable-node-id" "the argument is not a learnable node")
	  "")))

;; get-learnable-node-rule : learnable-node -> list symbols
;; A function that obtains a rule of a learnable node;
;; it returns an empty list if the argument is not valid.
;;
;; Note that learnable-node-rule is provided by Gambit

(define (get-learnable-node-loopy_rule learn)
   (if (is-learnable-node? learn)
       (learnable-node-loopy_rule learn)
       (begin
	  (er "get-learnable-node-rule" "the argument is not a learnable node")
	  (list))))

;; get-learnable-node-table : learnable-node -> matrix
;; A function that obtains a current approximation of a conditional probability
;; table of a learnable node; it returns an empty matrix if the argument
;; is not valid.
;;
;; Note that learnable-node-table is provided by Gambit

(define (get-learnable-node-table learn)
   (if (is-learnable-node? learn)
       (learnable-node-table learn)
       (begin
	  (er "get-learnable-node-table" "the argument is not a learnable node")
	  empty-loopy-matrix)))

;; get-learnable-node-neighbors : learnable-node -> list cluster-node
;; A function that obtains a list of neighbors of a learnable node;
;; it returns an empty list if the argument is not valid.
;;
;; Note that learnable-node-neighbors is provided by Gambit

(define (get-learnable-node-neighbors learn)
   (if (is-learnable-node? learn)
       (learnable-node-neighbors learn)
       (begin
	  (er "get-learnable-node-neighbors" "the argument is not a learnable node")
	  (list))))

;; the definition of a link structure that defines a connection between a variable and a cluster
;; id                     -- an ID of a link                                  (string)
;; variable               -- a variable node on one of the ends of a link     (variable-node)
;; cluster                -- a cluster node on one of the ends of a link      (cluster-node)
;; message-from-variable  -- a message from a variable node to a cluster node (vector)
;; message-from-cluster   -- a message from a cluster node to a variable node (vector)

(define-structure edge id variable cluster message-from-variable message-from-cluster)

;; a definition of a global counter of all links

(define edge-count 0)

;; the definitions of the empty link
;;
;; Note that make-edge is provided by Gambit

(define empty-edge (make-edge "" empty-variable-node empty-cluster-node empty-vector empty-vector))

;; initialize-message : integer -> vector
;; A function that makes an initial message vector for a link
;; (initially the message is set to a vector of ones);
;; it returns an empty vector if the argument is not a positive integer.

(define (initialize-message n)
   (if (integer? n)
       (if (> n 0)
	   (vector-of-ones n)
	   (begin
	      (er "initialize-message" "the argument is not a positive integer")
	      empty-vector))
       (begin
	  (er "initialize-message" "the argument is not an integer")
	  empty-vector)))

;; create-new-edge : variable-node, cluster-node -> edge
;; A function that creates a new link;
;; to return an empty link if the arguments are not valid.
;;
;; Note that make-edge is provided by Gambit

(define (create-new-edge var clus)
   (cond
      ((not (is-variable-node? var))(begin
				       (er "create-new-edge" "the first argument is not a variable node")
				       empty-edge))
      ((not (is-cluster-node? clus))(begin
				      (er "create-new-edge" "the second argument is not a cluster node")
				      empty-edge))
      (else (let ((n edge-count)(d (get-variable-node-dim var)))
	       (set! edge-count (+ edge-count 1))
	       (make-edge (make-id "_" n) var clus (initialize-message d) (initialize-message d))))))

;; is-edge? : object -> boolean
;; A function that returns true if an object is a valid link;
;; to return false if the object is not.
;;
;; Note that edge?, edge-id, edge-variable, edge-cluster, edge-message-from-variable, edge-message-from-cluster are provided by Gambit

(define (is-edge? o)
   (if (edge? o)
       (and (string? (edge-id o))
	    (is-variable-node? (edge-variable o))
	    (is-cluster-node? (edge-cluster o))
	    (vector? (edge-message-from-variable o))
	    (vector? (edge-message-from-cluster o)))
       #f))

;; get-edge-id : edge -> string
;; A function that obtains an ID of a link;
;; it returns an empty string if the argument is not a valid link.
;;
;; Note that edge-id is provided by Gambit

(define (get-edge-id lnk)
   (if (is-edge? lnk)
       (edge-id lnk)
       (begin
	  (er "get-edge-id" "the argument is not a link")
	  "")))

;; get-edge-variable : edge -> variable-node
;; A function that obtains a variable node connected to a link;
;; it returns an empty variable node if the argument is not a valid link.
;;
;; Note that edge-variable is provided by Gambit

(define (get-edge-variable lnk)
   (if (is-edge? lnk)
       (edge-variable lnk)
       (begin
	  (er "get-edge-variable" "the argument is not a link")
	  empty-variable-node)))

;; get-edge-cluster : edge -> cluster-node
;; A function that obtains a cluster node connected to a link;
;; it returns an empty cluster node if the argument is not a valid link.
;;
;; Note that edge-cluster is provided by Gambit

(define (get-edge-cluster lnk)
   (if (is-edge? lnk)
       (edge-cluster lnk)
       (begin
	  (er "get-edge-cluster" "the argument is not a link")
	  empty-cluster-node)))

;; get-edge-message-from-variable : edge -> vector
;; A function that obtains a message from the variable node to the cluster node;
;; it returns an empty vector if the argument is not a legal link.
;;
;; Note that edge-message-from-variable is provided by Gambit

(define (get-edge-message-from-variable lnk)
   (if (is-edge? lnk)
       (edge-message-from-variable lnk)
       (begin
	  (er "get-edge-message-from-variable" "the argument is not a link")
	  empty-vector)))

;; get-edge-message-from-cluster : edge -> vector
;; A function that obtains a message from the cluster node to the variable node;
;; it returns an empty vector if the argument is not a legal link.
;;
;; Note that edge-message-from-cluster is provided by Gambit

(define (get-edge-message-from-cluster lnk)
   (if (is-edge? lnk)
       (edge-message-from-cluster lnk)
       (begin
	  (er "get-edge-message-from-cluster" "the argument is not a link")
	  empty-vector)))

;; get-cluster-node-id : cluster-node -> string
;; A function that returns an ID of a cluster node;
;; it returns an empty string if the argument is not a valid cluster node.
;;
;; Note that cluster-node-id is provided by Gambit

(define (get-cluster-node-id cl)
   (if (is-cluster-node? cl)
       (cluster-node-id cl)
       (begin
	  (er "get-cluster-node-id" "the argument is not a cluster node")
	  "")))

;; get-cluster-node-table : cluster-node -> matrix
;; A function that returns an appropriate conditional probability table, depending on whether the cluster
;; is attached to a learnable node or not; it returns an empty table if the argument is not
;; a legal cluster node.
;;
;; Note that cluster-node-table, cluster-node-learnable, learnable-node-table are provided by Gambit

(define (get-cluster-node-table cl)
   (if (is-cluster-node? cl)
       (let ((l (cluster-node-learnable cl)))
	  (if (null? l)
	      ;; if the cluster node is not connected to a learnable node, we just return its table
	      (cluster-node-table cl)
	      ;; otherwise we take the current approximation from the learnable node and return its normalized version
	      (conditionally-normalize-matrix (learnable-node-table (car l)))))
       (begin
	  (er "get-cluster-node-table" "the argument is not a cluster node")
	  empty-loopy-matrix)))

;; get-cluster-node-potential : cluster-node -> matrix
;; A function that obtains a potential of a cluster node;
;; it returns an empty matrix if the argument is not legal.
;;
;; Note that cluster-node-potential is provided by Gambit

(define (get-cluster-node-potential cl)
      (if (is-cluster-node? cl)
       (cluster-node-potential cl)
       (begin
	  (er "get-cluster-node-potential" "the argument is not a cluster node")
	  empty-loopy-matrix)))

;; get-cluster-node-neighbors : cluster-node -> list edge
;; A function that retrieves a list of links to which a cluster node is connected;
;; if the argument is not valid, it returns an empty list.
;;
;; Note that cluster-node-neighbors is provided by Gambit

(define (get-cluster-node-neighbors cl)
      (if (is-cluster-node? cl)
       (cluster-node-neighbors cl)
       (begin
	  (er "get-cluster-node-neighbors" "the argument is not a cluster node")
	  (list))))

;; get-cluster-node-learnable : cluster-node -> list learnable-node
;; A function that gets the list possibly containing a learnable node of a cluster node;
;; it returns an empty list if the argument is not a cluster node.
;;
;; Note that cluster-node-learnable is provided by Gambit

(define (get-cluster-node-learnable cl)
      (if (is-cluster-node? cl)
       (cluster-node-learnable cl)
       (begin
	  (er "get-cluster-node-learnable" "the argument is not a cluster node")
	  (list))))


;; create-new-cluster-node : list variable-node, matrix -> cluster-node
;; A function that makes a cluster node;
;; it returns an empty cluster node if the arguments are not valid.

(define (create-new-cluster-node lst table)

   ;; make-edge-list : list variable-nodes, cluster-node -> list edge
   ;; this subroutine traverses a list of variable nodes and 
   ;; creates a link between the node and the cluster node, and returns
   ;; a list of these links.
   
   (define (make-edge-list lst clust)
      (if (is-cluster-node? clust)
	  (if (null? lst)
	      (list)
	      (let ((head (car lst)))
		 (if (is-variable-node? head)
		     (let ((l (create-new-edge head clust)))
			(variable-node-neighbors-set! head (cons l (variable-node-neighbors head)))  ;;!!!!!!!!!!!!!!!!! change to a set subroutine!!!!!!!!!!!
			(cons l (make-edge-list (cdr lst) clust)))
		     (begin
			(er "make-edge-list" "attempting to make a link with a NON variable node")
			(list)))))
	  (begin
	     (er "make-edge-list" "attempting to make a link with a NON cluster node")
	     (list))))

   ;; get-dimensions : list variable-node -> list integer
   ;; a subroutine that computes the list of dimensions from the list of variable nodes.

   (define (get-dimensions l)
      (if (null? l)
	  (begin
	     (er "get_dimensions" "the argument is an empty list")
	     (list))
	  (let loop ((l l)(acc '()))
	     (cond ((null? l) acc)
		   (else (loop (cdr l) (append acc (cons (get-variable-node-dim (car l)) '()))))))))

   ;; the body of the create-new-cluster-node function

   ;; construct a dimensions list
   (let* ((d (list->vector (get-dimensions lst)))
	  ;; create a cluster node
	  (c (make-cluster-node (make-id "C" cluster-node-count) table (initialize-potential d) '() '() #f)))
      ;; make a list of links between the cluster and variables, and assign it to the cluster node
      (cluster-node-neighbors-set! c (make-edge-list lst c))
      (set! cluster-node-count (+ cluster-node-count 1))
      c))

;; create-new-cluster-node-learnable : list variable-node -> cluster-node
;; A function that creates a cluster node in case when the rule it is associated with is learnable, so
;; there is no known matrix.

(define (create-new-cluster-node-learnable lst)
   (let loop ((l lst)(acc '()))
      (if (null? l)
	  (create-new-cluster-node lst (make-loopy-matrix (list->vector acc)))
	  (loop (cdr l) (append acc (list (get-variable-node-dim (car l))))))))

;; add-cluster-node-to-learnable-node : cluster-node, learnable-node -> void
;; A function that adds a cluster node to the neighbor list of the learnable node and to update the current approximation
;; of the table in the learnable node; or it yells about the error if the arguments are not valid.

(define (add-cluster-node-to-learnable-node clust learn)
   (cond
      ((not (is-cluster-node? clust))(er "add-cluster-node-to-learnable-node" "the first argument is not a cluster node"))
      ((not (is-learnable-node? learn))(er "add-cluster-node-to-learnable-node" "the second argument is not a learnable node"))
      ;; add the cluster to the neighbor list of the learnable
      (else (learnable-node-neighbors-set! learn (cons clust (get-learnable-node-neighbors learn)))
	    ;; mark cluster as to be learnt by putting learnable on its list
	    (cluster-node-learnable-set! clust (cons learn (get-cluster-node-learnable clust)))
	    ;; add a potential of the cluster to the current approximation of the table in learnable
	    (learnable-node-table-set! learn (matrix-plus-matrix (get-learnable-node-table learn) (get-cluster-node-potential clust))))))

;; the definition of a network structure
;; variables -- a list of all variable nodes  (list variable-node)
;; clusters  -- a list of all cluster nodes   (list cluster-node)
;; learnable -- a list of all learnable nodes (list learnable-nodes)

(define-structure network id last_tick variables clusters learnable)

;; a definition of a global counter of all networks

(define net-count 0)

;; the definitions of the empty network
;;
;; Note that make-network is provided by Gambit

(define empty-network (make-network "" (string->symbol (number->string 0)) '() '() '()))

;; create-new-network : list variable-node, list cluster-node, list learnable node -> network
;; A function that creates a new network;
;;
;; Note that make-network is provided by Gambit

(define (create-new-network var clust learn)
   (let ((n net-count))
      (set! net-count (+ net-count 1))
      (make-network (make-id "N" n) (string->symbol (number->string 0)) var clust learn)))

;; create a new empty network
(define initialized-network (create-new-network '() '() '()))

;; is-network? : object -> boolean
;; A function that returns true if an object is a valid network;
;; it returns false if the object is not.
;;
;; Note that network?, network-id are provided by Gambit

(define (is-network? o)
   (if (network? o)
       (string? (network-id o))  ;; as the network expands, this place will have more conditions !!!
       #f))

;; get-network-id : network -> string
;; A function that obtains an ID of a network; or if the object is not a network,
;; it returns an empty string.
;;
;; Note that network-id is provided by Gambit

(define (get-network-id o)
   (if (is-network? o)
       (network-id o)
       (begin
	  (er "get-network-id" "the argument is not a network")
	  "")))

;; get-network-variables : network -> list variable-node
;; A function that obtains a list of variable nodes of a network; or if the object is not a network,
;; it returns an empty list.
;;
;; Note that network-variables is provided by Gambit

(define (get-network-variables o)
   (if (is-network? o)
       (network-variables o)
       (begin
	  (er "get-network-variables" "the argument is not a network")
	  (list))))

;; get-network-clusters : network -> list cluster-node
;; A function that obtains a list of cluster nodes of a network; or if the object is not a network,
;; it returns an empty list.
;;
;; Note that network-clusters is provided by Gambit

(define (get-network-clusters o)
   (if (is-network? o)
       (network-clusters o)
       (begin
	  (er "get-network-clusters" "the argument is not a network")
	  (list))))

;; get-network-learnable : network -> list learnable-node
;; A function that obtains a list of learnable nodes of a network; or if the object is not a network,
;; it returns an empty list.
;;
;; Note that network-learnable is provided by Gambit

(define (get-network-learnable o)
   (if (is-network? o)
       (network-learnable o)
       (begin
	  (er "get-network-learnable" "the argument is not a network")
	  (list))))

;; add-node-to-network : network, object -> void
;; A function that adds a node to a network; if the object is not a node
;; it yells about an error.

(define (add-node-to-network net node)
   (if (is-network? net)
       (cond ((is-variable-node? node)(network-variables-set! net (cons node (get-network-variables net))))
	     ((is-cluster-node? node)(network-clusters-set! net (cons node (get-network-clusters net))))
	     ((is-learnable-node? node)(network-learnable-set! net (cons node (get-network-learnable net))))
	     (else (er "add-node-to-network" "the second argument is not a node")))
       (er "add-node-to-network" "the first argument is not a network")))

;; get-variable-node : list symbols, network -> variable-node
;; A function that returns a variable node from the network whose term is equal to a given term;
;; or it returns #f if the arguments are not valid.

(define (get-variable-node term net) ;+++++++++++++++++++++++ use HASH!!!!!!!!!!!!!!!!!!!!!
   (cond
      ((null? term)(begin
		      (er "get-variable-node" "the first argument is an empty list")
		      empty-variable-node))
      ((not (is-network? net))(begin
			     (er "get-variable-node" "the second argument is not a network")
			     empty-variable-node))
      (else (let ((lst (get-network-variables net)))
	       (let loop ((l lst))
		  (cond
		     ((null? l) #f)
		     ((equal? (get-variable-node-term (car l)) term) (car l))
		     (else (loop (cdr l)))))))))

;; get-learnable-node : list symbols, network -> learnable-node
;; A function that returns a learnable node from the network whose rule is equal to a given list;
;; or it returns #f if the arguments are not valid.

(define (get-learnable-node loopy_rule net)
   (cond
      ((not (rule? loopy_rule))(begin
				  (er "get-learnable-node" "the first argument is not a rule")
				  empty-learnable-node))
      ((not (is-network? net))(begin
				 (er "get-learnable-node" "the second argument is not a network")
				 empty-learnable-node))
      (else (let ((lst (get-network-learnable net)))
	       (let loop ((l lst))
		  (cond
		     ((null? l) #f)
		     ((eq? (get-learnable-node-loopy_rule (car l)) loopy_rule) (car l))
		     (else (loop (cdr l)))))))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;; printing out the network

;; print-node : object -> void
;; A function that prints out all the necessary information about the node;
;; or it spits the error out if the object is not a node.

(define (print-node node)
   ;; a subfunction to printout a list of links
   (define (print-edge-list lst)
      (if (null? lst)
	  (display "\n++++++++++++++++++++++\n")
	  (if (is-edge? (car lst))
	      (begin (display "\nLink: ")
		     (display (get-edge-id (car lst)))
		     (display "\nConnects a variable node ")
		     (display (get-variable-node-id (edge-variable (car lst))))
		     (display " with a cluster node ")
		     (display (get-cluster-node-id (edge-cluster (car lst))))
		     (display "\nSends from var to cl: ")
		     (print-vector (get-edge-message-from-variable (car lst)))
		     (display "\nSends from cl to var: ")
 		     (print-vector (get-edge-message-from-cluster (car lst)))
		     (print-edge-list (cdr lst)))
	      (er "print_edge_list" "the element of the list is not a link"))))
   ;; a subfunction to print out a list of clusters
   (define (print-cluster-node-list lst)
      (if (null? lst)
	  (display "\n++++++++++++++++++++++\n")
	  (if (is-cluster-node? (car lst))
	      (begin (display "\nCluster: ")
		     (display (get-cluster-node-id (car lst)))
		     (print-cluster-node-list (cdr lst)))
	      (er "print_cluster_list" "the element of the list is not a cluster"))))
   ;; a subfunction to print out a variable node
   (define (print-variable-node var)
      (if (is-variable-node? var)
	  (begin (display "\nVariable Node: ")
		 (display (get-variable-node-id var))
		 (display "\nTerm: ")
		 (let ((t (get-variable-node-term var)))
		   (pretty-print (cons (car t) (list (collapse-time (cadr t))))))
		 (display "Belief: ")
		 (print-vector (get-variable-node-belief var))
		 (display "\nLinks: ")
		 (print-edge-list (get-variable-node-neighbors var)))
	  (er "print_var" "the argument is not a variable node")))
   ;; a subfunction to print out a cluster node
   (define (print-cluster-node clust)
      (if (is-cluster-node? clust)
	  (begin (display "\nCluster Node: ")
		 (display (get-cluster-node-id clust))
		 (display "\nLearnable: ")
		 (let ((l (get-cluster-node-learnable clust)))
		    (if (null? l)
			(display "Empty")
			(display (get-learnable-node-id (car l)))))
		 (display "\nTable: ")
		 (print-matrix (get-cluster-node-table clust))
		 (display "\nPotential: ")
		 (print-matrix (get-cluster-node-potential clust))
		 (display "\nLinks: ")
		 (print-edge-list (get-cluster-node-neighbors clust)))
	  (er "print_clust" "the argument is not a cluster node")))
   ;; a subfunction to print out a learnable node
   (define (print-learnable-node learn)
      (if (is-learnable-node? learn)
	  (begin (display "\nLearnable Node: ")
		 (display (get-learnable-node-id learn))
		 (display "\nRule: ")
		 (pretty-print (cons (get-loopy-rule-head (learnable-node-loopy_rule learn))
				     (get-loopy-rule-body (learnable-node-loopy_rule learn))))
		 (display "Table: ")
		 (display "(unnormalized) ")
		 (print-matrix (get-learnable-node-table learn))
		 (display "\n       ")
		 (display "(normalized)   ")
		 (print-matrix (conditionally-normalize-matrix (get-learnable-node-table learn)))
		 (display "\nNeighbors: ")
		 (print-cluster-node-list (get-learnable-node-neighbors learn)))
	  (er "print_learn" "the argument is not a learnable node")))
   ;; the body of the print-node function
   (cond ((is-variable-node? node)(print-variable-node node))
	 ((is-cluster-node? node)(print-cluster-node node))
	 ((is-learnable-node? node)(print-learnable-node node))
	 (else (er "print-node" "the argument is not a node"))))

;; a function that recursively prints out a list of some nodes
(define (print-node-list lst)
   (if (null? lst)
       (display "\n")
       (begin (print-node (car lst))
	      (print-node-list (cdr lst)))))

;; print-network : network -> void
;; A function that prints out the network;
;; or it spits out an error if the argument is not a network.

(define (print-network n) 
   (if (is-network? n)(begin
			 (display ">>>>>>>>>NETWORK:\n")
			 (display ">>>VARIABLES: ")
			 (print-node-list (get-network-variables n))
			 (display ">>>CLUSTERS: ")
			 (print-node-list (get-network-clusters n))
			 (display ">>>LEARNABLES: ")
			 (print-node-list (get-network-learnable n)))
       (er "print-network" "the argument is not a network")))

