;;;;;
;;;;; This file contains a lexer, a wrapper for the LALR parser,
;;;;; and several functions supporting parsing.
;;;;; This file is connected with ll_lalr_parser.VER.scm
;;;;;


;; merge : list, list, number -> list
;; supporting function that merges two list in the appropriate way:
;; merge( [], [], 2 ) = []
;; merge( [1 2], [], 2 ) = '*error-merge*
;; merge( [], [3 5 4 6], 2 ) = '*error-merge*
;; merge( [1 2], [3 5 4 6], 2 ) = [1 3 5 2 4 6]

(define (merge l1 l2 len)
   (define (merge-acc small large i res)
      (cond ((null? small) (if (null? large)
			       res
			       '*error-merge*))
	    ((null? large) '*error-merge*)
	    (else (let loop ((l large)(iter i)(acc '()))
		     (cond ((null? l)(if (zero? iter)
					 (merge-acc (cdr small) l i (append res (cons (car small) acc)))
					 '*error-merge*))
			   ((zero? iter)(merge-acc (cdr small) l i (append res (cons (car small) acc)))) 
			   (else (loop (cdr l) (- iter 1) (append acc (list (car l))))))))))
   (merge-acc l1 l2 len '()))

;; add-queries-from-list : list -> void
;; a function that traverses a list and
;; adds every element on the list to the global list of queries (queries-list)

(define (add-queries-from-list lst)
   (if (not (null? lst))
       (begin
	  (set! queries-list (add-query (make-loopy-query (car lst)) queries-list))
	  (add-queries-from-list (cdr lst)))))

;; make-lexer : opened-file, error-handling -> token
;; The function reads in an appropriate number of characters and returns a corresponding
;; token, or *eoi* when the end of a file is reached.

(define (make-lexer input-file errorp)
   (lambda ()
      (letrec ((skip-spaces
		;; a function that skips tabulation and white spaces
		(lambda ()
		   (let loop ((c (peek-char input-file)))
		      (if (and (not (eof-object? c))
			       (or (char=? c #\space) (char=? c #\tab)))
			  (begin
			     (read-char input-file)
			     (loop (peek-char input-file)))))))   

	       (read-number
		;; a function that reads in a number as a sequence of digits and returns a string
		;; of these digits
		(lambda (l)
		   (let ((c (peek-char input-file)))
		      (if (and (not (eof-object? c)) (char-numeric? c))
			  (read-number (cons (read-char input-file) l))
			  (apply string (reverse l))))))

	       (read-float
		;; a function that reads in a number as a sequence of digits and "e-10" and returns a string
		;; of these digits
		(lambda (l)
		   (let ((c (peek-char input-file)))
		      (cond ((and (not (eof-object? c)) (char-numeric? c)) (read-float (cons (read-char input-file) l)))
			    ((char=? c #\e) (let ((e (read-char input-file))(next (peek-char input-file)))
					      (if (char=? next #\-)
						  (read-number (cons (read-char input-file) (cons e l)))
						  (apply string (reverse l)))))
			    (else (apply string (reverse l)))))))

	       (read-id
		;; a function that reads in an identification and returns it as a symbol
		(lambda (l)
		   (let ((c (peek-char input-file)))
		      (if (and (not (eof-object? c)) 
			       (or (char-alphabetic? c)
				   (char=? c #\_)
				   (char-numeric? c))) ;>>>> should be more relaxed (to accept "id_2")
			  (read-id (cons (read-char input-file) l))
			  (string->symbol (list->string (reverse l)))))))

	       (skip-to-line-end
		;; a function that skips everything until the end of the line
		(lambda ()
		   (let ((c (peek-char input-file)))
		      (if (and (not (eof-object? c)) (not (char=? c #\newline)))
			  (begin (read-char input-file) (skip-to-line-end)))))))
	 
	 ;; first skip spaces
	 (skip-spaces)
	 ;; then read the next token
	 (let loop ((c (read-char input-file)))
	    (cond
	       ((eof-object? c)      '*eoi*)
	       ((char=? c #\newline) 'NEWLINE)
	       ((char=? c #\+)       'PLUS)
	       ((char=? c #\-)       'MINUS)
	       ((char=? c #\,)       'COMMA)
	       ((char=? c #\{)       'LCURLY)
	       ((char=? c #\})       'RCURLY)
	       ((char=? c #\[)       'LBRACKET)
	       ((char=? c #\])       'RBRACKET)
	       ((char=? c #\()       'LPAREN)
	       ((char=? c #\))       'RPAREN)
	       ((char=? c #\|)       'GIVEN)
	       ((char=? c #\=)       'EQ)
	       ((char=? c #\?)       'QUESTION)
	       ;; used for internal commands (such as #trace)
	       ((char=? c #\#)       'POUND)
	       ;; <-
	       ((and (char=? c #\<) (char=? (peek-char input-file) #\-)) (read-char input-file)
									 'BELONGS)
	       ;; if we encountered a digit, then it is either an integer or a float
	       ((char-numeric? c) (let ((num (read-number (list c))))
				     (if (char=? (peek-char input-file) #\.)
					 (begin (read-char input-file)
						(let ((tail (read-float (list))))
						   (if (zero? (string-length tail))
						       (cons 'INT (string->number num))
						       (cons 'FLOAT (string->number (string-append num "." tail))))))
					 (cons 'INT (string->number num)))))
	       ;; if we encountered ".", then it is a float
	       ((and (char=? c #\.) (char-numeric? (peek-char input-file))) (let ((tail (read-float (list (read-char input-file)))))
									       (cons 'FLOAT (string->number (string-append "." tail)))))
	       ;; if we encountered a lower-case letter, then it is an ID or a command name ("trace")
	       ((and (char-alphabetic? c) (char-lower-case? c)) (let ((word (read-id (list c))))
								   (cond ((eq? word 'trace) 'TRACE)
									 ((eq? word 'successor) 'SUCCESSOR)
									 ((eq? word 'window_size) 'WINDOW_SIZE)
									 ((eq? word successor) (cons 'ID 'succ))
									 (else (cons 'ID  word)))))
	       ;; if we encountered an upper-case letter, then it is a variable
	       ((and (char-alphabetic? c) (char-upper-case? c)) (let ((word (read-id (list c))))
								   (cons 'VAR word)))
	       ;; if we encountered "%", then we skip the rest of the line (comments)
	       ((char=? c #\%) (skip-to-line-end)
			       (loop (read-char input-file)))
	       ;; otherwise we got an error
	       (else                 
		(errorp "PARSE ERROR : illegal character: " c)
		(skip-spaces)
		(loop (read-char input-file))))))))


(define parser-exit #f)

;; The parser function that opens the file, splits it on tokens (make-lexer)
;; and parses it (loopy-parser). "parser-exit" is used to return from the parser
;; if an error occurred.

(define parser
   (lambda (file-name)
      (call-with-current-continuation
       (lambda (k)
	  (letrec ((errorp (lambda args
			      (for-each display args) (newline)))
		   (p (open-input-file file-name))
		   (start (lambda ()
			     (loopy-parser (make-lexer p errorp) errorp))))
	     (set! parser-exit (lambda ex
				  (errorp "PARSE ERROR: " ex)
				  (k ex)))
	     (start)
	     (close-input-port p))))))
