;; A function that makes a copy of a list of constant immutable
;; objects.
;;
(define (copy-simple-list lst)
   (if (not (list? lst))
       (begin
	  (er "copy-simple-list" "the argument is not a list")
	  '*error-copy-simple-list*)
       (let loop ((l lst)(res '()))
	  (if (null? l)
	      res
	      (loop (cdr l) (append res (list (car l))))))))

;; A function that makes a copy of a vector of constant immutable
;; objects.
;;
(define (copy-simple-vector vec)
   (if (not (vector? vec))
       (begin
	  (er "copy-simple-vector" "the argument is not a vector")
	  '*error-copy-simple-vector*)
       (list->vector (vector->list vec))))

;; A function that makes a copy of a loopy matrix.
;;
(define (copy-matrix mat)
   (if (not (matrix? mat))
       (begin
	  (er "copy-matrix" "the argument is not a matrix")
	  '*error-copy-matrix*)
       (make-matrix (copy-simple-vector (matrix-dims mat))
		    (copy-simple-vector (matrix-data mat)))))

;; A function that makes a copy of a variable node structure.
;; Note that the function does not copy the list of connections.
;; 
(define (copy-variable-node var)
   (if (not (variable-node? var))
       (begin
	  (er "copy-variabe-node" "the argument is not a variable node")
	  '*error-copy-variable-node*)
       (let ((tick (variable-node-tick var)))
	  (if (list? tick)
	      (make-variable-node (string-copy (variable-node-id var))
				  (copy-simple-list (variable-node-term var))
				  (copy-simple-list tick)
				  (variable-node-dim var)
				  (copy-simple-vector (variable-node-belief var))
				  (list)
				  var)
	      (make-variable-node (string-copy (variable-node-id var))
				  (copy-simple-list (variable-node-term var))
				  tick
				  (variable-node-dim var)
				  (copy-simple-vector (variable-node-belief var))
				  (list)
				  var)))))

;; A function that makes a copy of a cluster node. Note that the function
;; does not copy node's connections including connections to a learnable node.
;;
(define (copy-cluster-node clust)
   (if (not (cluster-node? clust))
       (begin
	  (er "copy-cluster-node" "the argument is not a cluster node")
	  '*error-copy-cluster-node*)
       (let ((learn (cluster-node-learnable clust)))
	  (if (null? learn)
	      ;; if the cluster node is not connected to a learnable node, we just copy its table while copying the node
	      (make-cluster-node (string-copy (cluster-node-id clust))
				 (copy-matrix (cluster-node-table clust))
				 (copy-matrix (cluster-node-potential clust))
				 (list)
				 (list)
				 clust)
	      ;; otherwise we take the current approximation from the learnable node and copy it as a table of the cluster node
	      (make-cluster-node (string-copy (cluster-node-id clust))
				 (copy-matrix (conditionally-normalize-matrix (learnable-node-table (car learn))))
				 (copy-matrix (cluster-node-potential clust))
				 (list)
				 (list)
				 clust)))))

;; A function that makes a copy of an edge. Note that the function does not copy
;; a variable and a cluster of the edge.
;;
(define (copy-edge e)
   (if (not (edge? e))
       (begin
	  (er "copy-edge" "the argument is not an edge")
	  '*error-copy-edge*)
       (make-edge (string-copy (edge-id e))
		  #f
		  #f
		  (copy-simple-vector (edge-message-from-variable e))
		  (copy-simple-vector (edge-message-from-cluster e)))))


(define (find-copy orig vars)
   (cond ((not (variable-node? orig)) (er "find-copy" "the first argument is not a variable node")
				      #f)
	 ((null? vars) (er "find-copy" "the second argument is an empty list")
		       #f)
	 (else (let loop ((l vars))
		  (if (null? l)
		      #f
		      (let ((elem (car l)))
			 (if (eq? (variable-node-host elem) orig)
			     elem
			     (loop (cdr l)))))))))

(define (attach clust vars)
   (cond ((not (cluster-node? clust)) (er "attach" "the first argument is not a cluster node")
				      '*error-attach*)
	 ((null? vars) (er "attach" "the second argument is an empty list")
		       '*error-attach*)
	 (else
	  (let ((orig-clust (cluster-node-host clust)))
	     (let loop ((l (cluster-node-neighbors orig-clust)))
		(if (not (null? l))
		    (let* ((e (car l))
			   (new-edge (copy-edge e)))
		       (edge-cluster-set! new-edge clust)
		       (let ((var (find-copy (edge-variable e) vars)))
			  (if var
			      (begin
				 (edge-variable-set! new-edge var)
				 (variable-node-neighbors-set! var (cons new-edge (variable-node-neighbors var)));order?
				 (cluster-node-neighbors-set! clust (append (cluster-node-neighbors clust) (list new-edge)))
				 (loop (cdr l)))
			      (begin
				 (er "attach" "could not find a copy of a variable node")
				 '*error-attach*))))))))))


(define (copy-network var qnet)

   (define (subtract-lists minuend subtrahend)
      (let loop ((l minuend)(res (list)))
	 (if (null? l)
	     (reverse res)
	     (let ((head (car l))(tail (cdr l)))
		(if (memv head subtrahend)
		    (loop tail res)
		    (loop tail (cons head res)))))))

   (define (copy-cluster-neighbors neighbor-variables nt)
      (let loop ((l neighbor-variables)
		 (neighbor-variable-copies (list))
		 (first-encounter (list)))
	 (if (null? l)
	     (cons neighbor-variable-copies (list first-encounter))
	     (let* ((head (car l))
		    (copy (find-copy head (network-variables nt))))
		(if copy
		    (loop (cdr l)
			  (append neighbor-variable-copies (list copy))
			  first-encounter)
		    (let ((new-variable-copy (copy-variable-node head)))
		       (add-node-to-network nt new-variable-copy)
		       (loop (cdr l)
			     (append neighbor-variable-copies (list new-variable-copy))
			     (append first-encounter (list new-variable-copy)))))))))
   

   (cond
      ((not (is-variable-node? var)) (er "copy-network" "the first argument is not a variable node")
					'*error-argument*)
      ((not (is-network? qnet)) (er "copy-network" "the second argument is not a network")
				'*error-argument*)
      (else
       (let ((var-copy (copy-variable-node var)))
	  (add-node-to-network qnet var-copy)
	  (let loop ((first-encounter (list var-copy)))
	     (if (not (null? first-encounter))
		 (let* ((host-var (variable-node-host (car first-encounter)))
			(not-copied-cluster-neighbors (subtract-lists (map get-edge-cluster (get-variable-node-neighbors host-var))
								      (map cluster-node-host (network-clusters qnet)))))
		    (let iter ((l not-copied-cluster-neighbors))
		       (if (null? l)
			   (loop (cdr first-encounter))
			   (let ((var-neighbors (map get-edge-variable (get-cluster-node-neighbors (car l)))));;;;;;;;
			      (let* ((copy (copy-cluster-neighbors var-neighbors qnet))
				     (var-neighbor-copies (car copy))
				     (fresh-vars (cadr copy)))
				 (set! first-encounter (append first-encounter fresh-vars))
				 (let ((new-cluster (copy-cluster-node (car l))))
				    (attach new-cluster var-neighbor-copies)
				    (add-node-to-network qnet new-cluster)
				    (iter (cdr l))))))))))
	  var-copy))))




(define (construct-query-subnet-from-rule node rule qnet net)
   ;; A supporting function that takes a list of variable nodes and
   ;; constructs a cluster node connecting those; the function ignores
   ;; if the cluster learnable.
   ;;
   (define (connect-variables-to-clusters rule neighbors net)
      (let ((mat (get-loopy-rule-matrix rule)))
	 (if (symbol? mat)
	     ;; if the matrix is a learnable distribution
	     ;; create a new cluster node
	     (let* ((learned (rule-learned rule))
		    (clust (create-new-cluster-node neighbors learned)))
		;; add the cluster node to the network
		(add-node-to-network net clust))
	     ;; if the matrix is known, create a new cluster node and add it to the network
	     (let ((clust (create-new-cluster-node neighbors mat)))
		(add-node-to-network net clust)))))

   (cond ((not (is-variable-node? node)) '*error-construct-network*)
	 ((not (is-loopy-rule? rule)) '*error-construct-network*)
	 (else
	  ;; get the term of the variable node and try to unify it with the head of the rule
	  (let* ((term (get-variable-node-term node))(subs (unify term (get-loopy-rule-head rule))))
;	    (display "SUBSTITUTIONS:")(pretty-print subs);;;;;
	     (if (eq? subs '*failure*)
		 (list)
		 ;; iterate through the terms of the body of the rule after substitution
		 (let loop ((body (substitute-list subs (get-loopy-rule-body rule)))(neighbors '())(new-vars '()))
;		   (display "SUBST BODY:")(pretty-print body);;;;;
		    (if (null? body)
			;; if there are no more terms in the body left, connect the variable nodes with a cluster node
			(begin
			   (connect-variables-to-clusters rule
							  (cons node neighbors)
							  qnet)
			   ;; return newly created variable nodes
			   new-vars)
			;; try to get a variable node in the query-net corresponding to the term from the body
			(let* ((body-term (car body))
			       (body-qvar (get-variable-node body-term qnet)))
;			  (display "BODY-TERM:")(pretty-print body-term);;;;;
			   ;; if there is such a variable node, add it to the list of neighbors and iterate
			   (if body-qvar
			       (loop (cdr body) (append neighbors (list body-qvar)) new-vars)
			       ;; otherwise, check if there is a corresponding var node in the original net
			       (let ((body-var (get-variable-node body-term net)))
;				 (display "BODY-TERM (SECOND):")(pretty-print body-term);;;;;
;				 (display "BODY-VAR:")(pretty-print (if body-var (cons (variable-node-id body-var)(list (variable-node-term body-var))) 'NONE));;;;;;;;
				  (if body-var
				      ;; if there is, make a copy of the network starting from the node and add the copied node to the list
				      (let ((copy-body-var (copy-network body-var qnet)))
;					(display "EXISTING VAR?:")(pretty-print (cons (variable-node-id body-var)(list (variable-node-term body-var))));;;;;;;;;;;;;
;					(display "CONNECTING TO EXISTING VAR:")(pretty-print (cons (variable-node-id copy-body-var)(list (variable-node-term copy-body-var))));;;;;;;;;;;;;
					 (loop (cdr body) (append neighbors (list copy-body-var)) new-vars))
				      ;; otherwise create a new variable node, add it to the list of neighbors, and iterate
				      (let ((nvar (create-new-variable-node body-term
									    (get-loopy-definition-dim (definitions-hash 'get-first-entry (car body-term))))))
					 (add-node-to-network qnet nvar)
					 (loop (cdr body) (append neighbors (list nvar)) (append new-vars (list nvar))))))
				  )))))))))


(define (complete-query-net-from-vars lst qnet net)
;  (display "\nCOMPLETING NET FROM VARS\n");;;;;
   (let loop ((current lst))
      (if (not (null? current))
	  ;; take the first variable node on the list
	  (let ((var-node (car current)))
	     ;; check if this is a variable node
;	    (display "CURRENT LIST: ")(pretty-print (map variable-node-term current));;;
;	    (display "FIRST VAR: ")(pretty-print (cons (variable-node-id var-node)(list (variable-node-term var-node))));;;
	     (if (not (is-variable-node? var-node))
		 (begin
		    (er "complete-query-net-from-vars" "the argument is not a list of variable nodes")
		    '*error-construct-query-network*)
		 ;; iterate through all the rules
		 (let iter-rules ((r (rules-hash 'get-every-entry (car (get-variable-node-term var-node))))
				  (extended-current (cdr current)))
;		   (display "RULES: ")(pretty-print r);;;;;;
;		   (display "EXTENDED LIST: ")(pretty-print (map variable-node-term extended-current));;;
		    (if (or (null? r) (eq? r '*hash-table-search-failed*))
			;; once we checked every rule, take another variable node from the extended list
			(loop extended-current)
			;; for a rule on the list, construct a subnetwork
			(let ((res (construct-query-subnet-from-rule var-node (car r) qnet net)))
;			  (display "RES RULE-NET LIST: ")(pretty-print (map variable-node-term res));;;
			   (if (eq? res '*error-construct-query-network*)
			       '*error-construct-query-network*
			       ;; add new variable nodes on the list without repetition
			       (iter-rules (cdr r) (merge-lists extended-current res)))))))))))


(define (complete-query-net-from-query q current-net query-net)
   ;; try to find a variable node corresponding to a term in a query-net
   (let* ((qterm (get-loopy-query-term q))(qvar (get-variable-node qterm query-net)))
;     (display "\nCOMPLETING NET FROM QUERY\n");;;;;
      (if (not qvar)
	  ;; if no such variable node is found, check if there is such a node in the original network
	  (let ((var (get-variable-node qterm current-net)))
	     (if var
		 (copy-network var query-net)
		 (let ((new-var (create-new-variable-node qterm
							  (get-loopy-definition-dim (definitions-hash 'get-first-entry (car qterm))))))
;		   (display "NEW VAR: ")(pretty-print (cons (variable-node-id new-var)(list qterm)));;;;;;;
		    (add-node-to-network query-net new-var)
		    ;; complete the network for the new variable node
		    (complete-query-net-from-vars (list new-var) query-net current-net)
		    ))))))

;; what if the node is not in the query-net but in the current net???
;; is it possible??
;;
(define (construct-query-net-from-fact f current-net query-net)
   ;; try to find a variable node in the query net corresponding to a term
   (let* ((fterm (get-loopy-fact-term f))(fvar (get-variable-node fterm query-net)))
      (if fvar
	  ;; if such a variable node exists, simply create a leaf cluster node
	  (let ((c (create-new-cluster-node (list fvar) (get-loopy-fact-value f))))
	     (add-node-to-network query-net c))
	  ;; otherwise create a new variable node
	  (let ((new-var (create-new-variable-node fterm
						   (get-loopy-definition-dim (definitions-hash 'get-first-entry (car fterm))))))
	     (add-node-to-network query-net new-var)
	     (add-node-to-network query-net (create-new-cluster-node (list new-var) (get-loopy-fact-value f)))
	     ;; complete the network for the new variable node
	     (complete-query-net-from-vars (list new-var) query-net current-net)))))

(define (make-query-net current-time current-net) 
   (define (convert-to-number term)
      (cond ((eq? term (string->symbol (number->string 0))) 0)
	    ((and (list? term) (not (null? term)) (eq? (car term) 'succ)) (+ 1 (convert-to-number (cadr term))))
	    (else (er "convert-to-number" "the argument is empty")
		  '*error-argument*)))

   (define (make-relative-query q t)
      (let* ((qterm (query-term q))
	     (qname (car qterm))
	     (qtime (convert-to-number (cadr qterm)))
	     (time (convert-to-number t))
	     (diff (- qtime time)))
	 (cond ((> diff 0) (let loop ((n diff)(res 't))
			      (if (= n 0)
				  (begin
				     (query-term-set! q (cons qname (list res)))
				     q)
				  (loop (- n 1) (cons 'succ (list res))))))
	       ((< diff 0) (let loop ((n diff)(res 't))
			      (if (= n 0)
				  (begin
				     (query-term-set! q (cons qname (list res)))
				     q)
				  (loop (+ n 1) (cons 'pred (list res))))))
	       (else (query-term-set! q (cons qname (list 't)))
		     q))))

   (define (make-relative-fact-depending-on-query f q)
      (let ((fact-term-name (car (fact-term f)))
	    (query-time (cdr (query-term q))))
	 (fact-term-set! f (cons fact-term-name query-time))
	 f))

   (define (get-facts-from-same-time-as-query time-term)
      (let* ((dat cosmos-data)
	     (cur (get-current-facts dat)))
	 (if (null? cur)
	     '()
	     (if (equal? (cadr (fact-term (car cur))) time-term)
		 cur
		 (let loop ((l (get-next-given-facts dat)))
		    (if (null? l)
			'()
			(if (equal? (cadr (fact-term (cadr l))) time-term)
			    l
			    (loop (get-next-given-facts dat)))))))))
   
   (let ((query-net (create-new-network (list) (list) (list))))
;      (display "QQQNET\n");;;
      (let loop ((l queries-list))
;	 (display "LOOP QUERIES: ")(pretty-print l);;;
	 (if (null? l)
	     query-net
	     (let* ((elem (car l))
		    (time-term (cdr (query-term elem)))
		    (facts (get-facts-from-same-time-as-query (car time-term)))
		    (rel-query (make-relative-query elem current-time)))
;		(display "TIME-TERM: ")(pretty-print time-term);;
;		(display "FACTS: ")(pretty-print facts);;
;		(display "REL-QUERY: ")(pretty-print rel-query);;
		(complete-query-net-from-query rel-query current-net query-net)
		(if (not (null? facts))
		    (let iter ((facts (cdr facts)))
;		      (display "ITER FACTS: ")(pretty-print facts);;
		      (if (not (null? facts))
			  (let ((fact (make-relative-fact-depending-on-query (car facts) rel-query)))
;			    (display "FACT: ")(pretty-print fact);;
;			    (display "CONSTR Q-NET FROM FACT\n");;;
			    (construct-query-net-from-fact fact current-net query-net)
			    (iter (cdr facts))))));;;;;;;;;analog of construct-network-from-fact
		(loop (cdr l)))))))

